"""
Testing for TG2 Configuration
"""
import os
import sys

import pytest
from sqlalchemy.engine import Engine
from sqlalchemy.orm import scoped_session, sessionmaker

try:
    import ming
except ImportError:
    # Ming is not supported on Python 3.8 
    ming = None
else:
    import ming.odm

from webtest import TestApp

import tg.i18n
from tests.base import setup_session_dir, teardown_session_dir, utcnow
from tg import (
    MinimalApplicationConfigurator,
    TGController,
    abort,
    expose,
    request,
    response,
)
from tg.appwrappers.mingflush import MingApplicationWrapper
from tg.configuration import config, milestones
from tg.configuration.app_config import AppConfig
from tg.configuration.auth import _AuthenticationForgerPlugin
from tg.configuration.auth.metadata import _AuthMetadataAuthenticator
from tg.configuration.tgconfig import _init_default_global_config
from tg.configuration.utils import TGConfigError, coerce_config, coerce_options
from tg.configurator import ApplicationConfigurator, FullStackApplicationConfigurator
from tg.configurator.base import (
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
    Configurator,
)
from tg.configurator.components.app_globals import AppGlobalsConfigurationComponent
from tg.configurator.components.auth import SimpleAuthenticationConfigurationComponent
from tg.configurator.components.caching import CachingConfigurationComponent
from tg.configurator.components.dispatch import DispatchConfigurationComponent
from tg.configurator.components.helpers import HelpersConfigurationComponent
from tg.configurator.components.i18n import I18NConfigurationComponent
from tg.configurator.components.ming import MingConfigurationComponent
from tg.configurator.components.paths import PathsConfigurationComponent
from tg.configurator.components.registry import RegistryConfigurationComponent
from tg.configurator.components.rendering import TemplateRenderingConfigurationComponent
from tg.configurator.components.session import SessionConfigurationComponent
from tg.configurator.components.sqlalchemy import SQLAlchemyConfigurationComponent
from tg.configurator.components.transactions import (
    TransactionManagerConfigurationComponent,
)
from tg.renderers.base import RendererFactory
from tg.support.converters import asbool, asint
from tg.util import Bunch
from tg.wsgiapp import TGApp


def setup_module():
    milestones._reset_all()
    setup_session_dir()


def teardown_module():
    milestones._reset_all()
    teardown_session_dir()


def _reset_global_config():
    milestones._reset_all()
    try:
        config.config_proxy.pop_thread_config()
    except:
        pass
    try:
        config.config_proxy.pop_process_config()
    except:
        pass


class PackageWithModel:
    __name__ = 'tests'
    __file__ = __file__

    def __init__(self):
        self.model = self.ModelClass()
        self.model.DBSession = self.model.FakeDBSession()

    class ModelClass:
        class FakeDBSession:
            def remove(self):
                self.DBSESSION_REMOVED=True

        def init_model(self, engine):
            if isinstance(engine, Engine):
                # SQLA
                return self.DBSession
            else:
                # Ming
                return dict(ming=True)

    class lib:
        class app_globals:
            class Globals:
                pass
PackageWithModel.__name__ = 'tests'


class UncopiableList(list):
    """
    This is to test configuration methods that make a copy
    of a list to modify it, using this we can check how it has
    been modified
    """
    def __copy__(self):
        return self

class FakeTransaction:
    def get(self):
        return self

    def begin(self):
        self.aborted = False
        self.doomed = False
        return self

    def abort(self):
        self.aborted = True

    def commit(self):
        self.aborted = False

    def _retryable(self, *args):
        return True
    note = _retryable

    def isDoomed(self):
        return self.doomed

    def doom(self):
        self.doomed = True

from tg.configuration.auth import TGAuthMetadata


class ApplicationAuthMetadata(TGAuthMetadata):
    def get_user(self, identity, userid):
        return {'name':'None'}

class ApplicationAuthMetadataWithAuthentication(TGAuthMetadata):
    def authenticate(self, environ, identity):
        return 1
    def get_user(self, identity, userid):
        return {'name':'None'}

class AtExitTestException(Exception):
    pass


class RootController(TGController):
    @expose()
    def test(self):
        return 'HI!'


class TestPylonsConfigWrapper:

    def setup_method(self):
        _reset_global_config()
        _init_default_global_config()
        self.config = config

    def teardown_method(self):
        _reset_global_config()
        _init_default_global_config()

    def test_create(self):
        pass

    def test_getitem(self):
        expected_keys = ['debug', 'package', 'tg.app_globals', 'tg.strict_tmpl_context']
        for key in expected_keys:
            self.config[key]

    def test_repr(self):
        _reset_global_config()
        assert repr(self.config) == '<TGConfig: missing>'
        _init_default_global_config()
        assert repr(self.config) == repr(self.config.config_proxy.current_conf())

    def test_getitem_bad(self):
        with pytest.raises(KeyError):
            self.config['no_such_key']

    def test_setitem(self):
        self.config['no_such_key'] = 'something'

    def test_delattr(self):
        del self.config.debug
        assert hasattr(self.config, 'debug') == False
        self.config.debug = False

    def test_delattr_bad(self):
        with pytest.raises(AttributeError):
            del self.config.i_dont_exist

    def test_keys(self):
        k = self.config.keys()
        assert 'tg.app_globals' in k


def test_coerce_config():
    opts = {'ming.connection.max_pool_size': '5'}
    conf = coerce_config(opts, 'ming.connection.', {'max_pool_size':asint})
    assert conf['max_pool_size'] == 5
    assert opts['ming.connection.max_pool_size'] == '5'


def test_coerce_options():
    opts = {'connection': 'false'}
    conf = coerce_options(opts, {'connection': asbool})
    assert conf['connection'] is False
    assert opts['connection'] == 'false'


class TestConfigurator:
    def setup_method(self):
        _reset_global_config()

    def teardown_method(self):
        _reset_global_config()
        tg.hooks._clear()  # Reset hooks

    def test_repr_action(self):
        act = BeforeConfigConfigurationAction()
        assert repr(act) == "<BeforeConfigConfigurationAction: None>"

    def test_reqlocal_configuration_dictionary(self):
        cfg = FullStackApplicationConfigurator()

        cfg.update_blueprint({'RANDOM_VALUE': 5})
        conf = cfg.configure({}, {})

        assert config['RANDOM_VALUE'] == 5
        assert len(config) == len(conf)

    def test_blueprint_invalid_view(self):
        cfg = FullStackApplicationConfigurator()
        try:
            cfg.get_blueprint_view('this.that.')
        except ValueError as e:
            assert str(e) == 'A Blueprint key cannot end with a .'
        else:
            assert False, 'Should have raised'

    def test_invalid_component(self):
        cfg = FullStackApplicationConfigurator()
        try:
            cfg.register(str)
        except ValueError as e:
            assert str(e) == 'Configuration component must inherit ConfigurationComponent'
        else:
            assert False, 'Should have raised'

    def test_replace_component(self):
        cfg = FullStackApplicationConfigurator()

        class TestComponentFirst(ConfigurationComponent):
            id = 'TESTCOMPONENT'

        class TestComponentSecond(ConfigurationComponent):
            id = 'TESTCOMPONENT2'

        cfg.register(TestComponentFirst)
        try:
            cfg.replace(TestComponentFirst, str)
        except ValueError as e:
            assert str(e) == 'Configuration component must inherit ConfigurationComponent'
        else:
            assert False, 'Should have raised'

        cfg.replace('TESTCOMPONENT', TestComponentSecond)
        comp = cfg.get_component('TESTCOMPONENT')
        assert isinstance(comp, TestComponentSecond), comp

    def test_component_without_id(self):
        cfg = FullStackApplicationConfigurator()

        class TestComponentFirst(ConfigurationComponent):
            pass

        try:
            cfg.register(TestComponentFirst)
        except ValueError as e:
            assert str(e).startswith('ConfigurationComponent must provide an id class attribute')
        else:
            assert False, 'Should have raised'

        try:
            cfg.replace(TestComponentFirst, TestComponentFirst)
        except ValueError as e:
            assert str(e).startswith('ConfigurationComponent must provide an id class attribute')
        else:
            assert False, 'Should have raised'

    def test_retrieve_current_configurator(self):
        cfg = FullStackApplicationConfigurator()
        cfg.update_blueprint({'RANDOM_VALUE': 5})
        cfg.configure({}, {})

        configurator = FullStackApplicationConfigurator.current()
        assert configurator.get_blueprint_value('RANDOM_VALUE') == 5

    def test_application_wrapper_replacement(self):
        class AppWrapperTest(object):
            def __init__(self, *args, **kwargs): pass
            def __call__(self, *args, **kw):
                return tg.Response('AppWrapper #1')

        class AppWrapperTestReplacement(object):
            def __init__(self, *args, **kwargs): pass
            def __call__(self, *args, **kw):
                return tg.Response('AppWrapper #2')

        cfg = FullStackApplicationConfigurator()
        cfg.update_blueprint({'root_controller': Bunch(index=lambda *args, **kwargs: 'HI')})
        cfg.register_application_wrapper(AppWrapperTest)

        app = TestApp(cfg.make_wsgi_app({'debug': True}, {}))
        assert app.get('/').text == 'AppWrapper #1', app.get('/').text

        cfg.replace_application_wrapper('AppWrapperTest', AppWrapperTestReplacement)

        app = TestApp(cfg.make_wsgi_app({}, {}))
        assert app.get('/').text == 'AppWrapper #2', app.get('/').text

    def test_sa_auth_requires_app_config(self):
        configurator = Configurator()
        configurator.register(SimpleAuthenticationConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e) == 'Simple Authentication only works on an ApplicationConfigurator'
        else:
            assert False, 'Should have raised'

    def test_sa_auth_authmetadata_without_authenticate(self):
        cfg = FullStackApplicationConfigurator()
        class FakeAuthMetadata():
            authenticate = None
        cfg.update_blueprint({
            'root_controller': Bunch(index=lambda *args, **kwargs: 'HI'),
            'sa_auth.enabled': True,
            'sa_auth.authmetadata': FakeAuthMetadata(),
            'sa_auth.cookie_secret': 'SECRET!'
        })
        cfg.make_wsgi_app({}, {})

    def test_caching_required_app_config(self):
        configurator = Configurator()
        configurator.register(CachingConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e) == 'Caching only works on an ApplicationConfigurator'
        else:
            assert False, 'Should have raised'

    def test_i18n_required_app_config(self):
        configurator = Configurator()
        configurator.register(I18NConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e) == 'I18N only works on an ApplicationConfigurator'
        else:
            assert False, 'Should have raised'

    def test_ming_required_app_config(self):
        configurator = Configurator()
        configurator.register(MingConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e).endswith('only works on an ApplicationConfigurator')
        else:
            assert False, 'Should have raised'

    def test_session_required_app_config(self):
        configurator = Configurator()
        configurator.register(SessionConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e).endswith('only work on an ApplicationConfigurator')
        else:
            assert False, 'Should have raised'

    def test_sqlalchemy_required_app_config(self):
        configurator = Configurator()
        configurator.register(SQLAlchemyConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e).endswith('only works on an ApplicationConfigurator')
        else:
            assert False, 'Should have raised'

    def test_transaction_required_app_config(self):
        configurator = Configurator()
        configurator.register(TransactionManagerConfigurationComponent)

        try:
            configurator.configure({}, {})
        except TGConfigError as e:
            assert str(e).endswith('only works on an ApplicationConfigurator')
        else:
            assert False, 'Should have raised'

    def test_dispatch_without_mimetypes(self):
        # This is exactly like MinimalApplicationConfigurator
        # but without the mimetypes component.
        apc = ApplicationConfigurator()
        apc.register(PathsConfigurationComponent, after=False)
        apc.register(DispatchConfigurationComponent, after=False)
        apc.register(AppGlobalsConfigurationComponent)
        apc.register(HelpersConfigurationComponent)
        apc.register(TemplateRenderingConfigurationComponent)
        apc.register(RegistryConfigurationComponent, after=True)

        class MinimalController(TGController):
            @expose()
            def index(self):
                return 'HI'

        apc.update_blueprint({
            'root_controller': MinimalController()
        })
        app = TestApp(apc.make_wsgi_app({}, {}))
        assert app.get('/').text == 'HI'

    def test_app_without_controller(self):
        cfg = MinimalApplicationConfigurator()
        app = TestApp(cfg.make_wsgi_app({}, {}))

        try:
            app.get('/')
        except TGConfigError as e:
            assert str(e) == 'Unable to load controllers, no controllers path configured!'
        else:
            assert False, 'Should have raised.'

    def test_tgapp_caches_controller_classes(self):
        class RootController(TGController):
            @expose()
            def index(self):
                return 'HI'

        tgapp = Bunch(app=None)
        def save_app(app):
            tgapp.app = app
            return app

        cfg = MinimalApplicationConfigurator()
        app = TestApp(cfg.make_wsgi_app({}, {}, wrap_app=save_app))

        tgapp.app.controller_classes['root'] = RootController
        assert app.get('/').text == 'HI'


class TestAppConfig:
    def setup_method(self):
        _reset_global_config()
        self.fake_package = PackageWithModel

    def teardown_method(self):
        _reset_global_config()
        tg.hooks._clear()  # Reset hooks

    def test_get_value(self):
        conf = AppConfig(minimal=True)
        conf['existing_value'] = 5
        assert conf['existing_value'] == 5
        assert conf.get('non_existing_value') == None

    def test_missing_attribute(self):
        conf = AppConfig(minimal=True)
        conf['existing_value'] = 5
        assert conf['existing_value'] == 5
        assert conf.existing_value == 5

        try:
            conf['missing_value']
        except KeyError:
            pass
        else:
            raise RuntimeError('Should have raised KeyError')

        try:
            conf.missing_value
        except AttributeError:
            pass
        else:
            raise RuntimeError('Should have raised AttributeError')

    def test_lang_can_be_changed_by_ini(self):
        conf = AppConfig(minimal=True)
        conf.make_wsgi_app(**{'i18n.lang': 'ru'})
        assert config['i18n.lang'] == 'ru'

    def test_create_minimal_app(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_create_minimal_app_with_factory(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        app_factory = conf.setup_tg_wsgi_app()
        app = app_factory()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_minimal_app_with_sqlalchemy(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        DBSession = scoped_session(sessionmaker(autoflush=True, autocommit=False))
        def init_model(engine):
            DBSession.configure(bind=engine)

        conf = AppConfig(minimal=True, root_controller=RootController())

        conf['use_sqlalchemy'] = True
        conf['sqlalchemy.url'] = 'sqlite://'
        conf['model'] = Bunch(DBSession=DBSession,
                              init_model=init_model)

        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_sqlalchemy_without_models(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['use_sqlalchemy'] = True
        conf['sqlalchemy.url'] = 'sqlite://'

        with pytest.raises(TGConfigError):
            app = conf.make_wsgi_app()

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_minimal_app_with_ming(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        mainsession = ming.Session()
        DBSession = ming.odm.ThreadLocalORMSession(mainsession)

        def init_model(engine):
            mainsession.bind = engine

        conf = AppConfig(minimal=True, root_controller=RootController())

        conf['use_ming'] = True
        conf['ming.url'] = 'mim:///dbname'
        conf['model'] = Bunch(init_model=init_model, DBSession=DBSession)

        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_ming_without_models(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        DBSession = scoped_session(sessionmaker(autoflush=True, autocommit=False))
        def init_model(engine):
            DBSession.configure(bind=engine)

        conf = AppConfig(minimal=True, root_controller=RootController())

        conf['use_ming'] = True
        conf['ming.url'] = 'mim://'

        with pytest.raises(TGConfigError):
            app = conf.make_wsgi_app()

    def test_setup_jinja_without_package(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.renderers = ['jinja']
        app = conf.make_wsgi_app()

    def test_setup_sqlalchemy(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf['sqlalchemy.url'] = 'sqlite://'
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert package.model.DBSession.DBSESSION_REMOVED

    def test_sqlalchemy_commit_veto(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

            @expose()
            def crash(self):
                raise Exception('crash')

            @expose()
            def forbidden(self):
                response.status = 403
                return 'FORBIDDEN'

            @expose()
            def notfound(self):
                response.status = 404
                return 'NOTFOUND'

        def custom_commit_veto(environ, status, headers):
            if status.startswith('404'):
                return True
            return False

        fake_transaction = FakeTransaction()
        import transaction
        prev_transaction_manager = transaction.manager
        transaction.manager = fake_transaction

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['package'] = package
        conf['model'] = package.model
        conf['use_sqlalchemy'] = True
        conf['tm.enabled'] = True
        conf['tm.commit_veto'] = custom_commit_veto
        conf['sqlalchemy.url'] = 'sqlite://'

        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert hasattr(conf, 'use_transaction_manager') is False

        app.get('/test')
        assert fake_transaction.aborted == False

        try:
            app.get('/crash')
        except:
            pass
        assert fake_transaction.aborted == True

        app.get('/forbidden', status=403)
        assert fake_transaction.aborted == False

        app.get('/notfound', status=404)
        assert fake_transaction.aborted == True

        transaction.manager = prev_transaction_manager

    def test_sqlalchemy_doom(self):
        fake_transaction = FakeTransaction()
        import transaction
        prev_transaction_manager = transaction.manager
        transaction.manager = fake_transaction

        class RootController(TGController):
            @expose()
            def test(self):
                fake_transaction.doom()
                return 'HI!'

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf['tm.enabled'] = True
        conf['sqlalchemy.url'] = 'sqlite://'

        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert hasattr(conf, 'use_transaction_manager') is False

        app.get('/test')
        assert fake_transaction.aborted == True

        transaction.manager = prev_transaction_manager

    def test_sqlalchemy_retry(self):
        fake_transaction = FakeTransaction()
        import transaction
        prev_transaction_manager = transaction.manager
        transaction.manager = fake_transaction

        from transaction.interfaces import TransientError

        class RootController(TGController):
            attempts = []

            @expose()
            def test(self):
                self.attempts.append(True)
                if len(self.attempts) == 3:
                    return 'HI!'
                raise TransientError()

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf['tm.enabled'] = True
        conf['sqlalchemy.url'] = 'sqlite://'
        conf['tm.attempts'] = 3

        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert hasattr(conf, 'use_transaction_manager') is False

        resp = app.get('/test')
        assert 'HI' in resp

        transaction.manager = prev_transaction_manager

    def test_setup_sqla_persistance(self):
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['sqlalchemy.url'] = 'sqlite://'
        conf.use_sqlalchemy = True
        conf.package = PackageWithModel()

        conf.make_wsgi_app()

    def test_setup_sqla_balanced(self):
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['sqlalchemy.master.url'] = 'sqlite://'
        conf['sqlalchemy.slaves.slave1.url'] = 'sqlite://'
        conf.use_sqlalchemy = True
        conf.package = PackageWithModel()

        conf.make_wsgi_app()

    def test_setup_sqla_balanced_prevent_slave_named_master(self):
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['sqlalchemy.master.url'] = 'sqlite://'
        conf['sqlalchemy.slaves.master.url'] = 'sqlite://'
        conf.use_sqlalchemy = True
        conf.package = PackageWithModel()

        with pytest.raises(TGConfigError):
            conf.make_wsgi_app()

    def test_setup_sqla_balanced_no_slaves(self):
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['sqlalchemy.master.url'] = 'sqlite://'
        conf.use_sqlalchemy = True
        conf.package = PackageWithModel()

        with pytest.raises(TGConfigError):
            conf.make_wsgi_app()

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mim://'
        conf['ming.db'] = 'inmemdb'

        app = conf.make_wsgi_app()

        tgapp = app.application
        while not isinstance(tgapp, TGApp):
            tgapp = tgapp.app

        ming_handler = tgapp.wrapped_dispatch
        while ming_handler != tgapp._dispatch:
            if isinstance(ming_handler, MingApplicationWrapper):
                break
            ming_handler = ming_handler.next_handler
        assert isinstance(ming_handler, MingApplicationWrapper), ming_handler

        class FakeMingSession(object):
            actions = []

            def flush_all(self):
                self.actions.append('FLUSH')

            def close_all(self):
                self.actions.append('CLOSE')

        ming_handler.ThreadLocalODMSession = FakeMingSession()

        app = TestApp(app)
        resp = app.get('/test')
        assert 'HI' in resp

        assert ming_handler.ThreadLocalODMSession.actions == ['FLUSH']

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_closes_on_failure(self):
        class RootController(TGController):
            @expose()
            def test(self):
                raise Exception('CRASH!')

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mim://'
        conf['ming.db'] = 'inmemdb'

        app = conf.make_wsgi_app()

        tgapp = app.application
        while not isinstance(tgapp, TGApp):
            tgapp = tgapp.app

        ming_handler = tgapp.wrapped_dispatch
        while ming_handler != tgapp._dispatch:
            if isinstance(ming_handler, MingApplicationWrapper):
                break
            ming_handler = ming_handler.next_handler
        assert isinstance(ming_handler, MingApplicationWrapper), ming_handler

        class FakeMingSession(object):
            actions = []

            def flush_all(self):
                self.actions.append('FLUSH')

            def close_all(self):
                self.actions.append('CLOSE')

        ming_handler.ThreadLocalODMSession = FakeMingSession()

        app = TestApp(app)

        try:
            app.get('/test', status=500)
        except:
            assert ming_handler.ThreadLocalODMSession.actions == ['CLOSE']
        else:
            assert False, 'Should have raised exception'

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_with_url_alone(self):
        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mim://inmemdb'

        app = conf.make_wsgi_app()
        assert app is not None

        dstore = config['tg.app_globals'].ming_datastore
        dstore_name = dstore.name
        # Looks like ming has empty dstore.name when using MIM.
        assert dstore_name == '', dstore

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_sqla_and_ming_both(self):
        package = PackageWithModel()
        base_config = AppConfig(minimal=True, root_controller=None)
        base_config.package = package
        base_config.model = package.model
        base_config.use_ming = True
        base_config['ming.url'] = 'mim://inmemdb'
        base_config.use_sqlalchemy = True
        base_config['sqlalchemy.url'] = 'sqlite://'

        app = base_config.make_wsgi_app()
        assert app is not None

        assert config['MingSession'], config
        assert config['tg.app_globals'].ming_datastore, config['tg.app_globals']

        assert config['SQLASession'], config
        assert config['tg.app_globals'].sa_engine, config['tg.app_globals']

        assert config['DBSession'] is config['SQLASession'], config

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_with_url_and_db(self):
        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mim://inmemdb'
        conf['ming.db'] = 'realinmemdb'

        app = conf.make_wsgi_app()
        assert app is not None

        dstore = config['tg.app_globals'].ming_datastore
        dstore_name = dstore.name
        assert dstore_name == 'realinmemdb', dstore

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_advanced_options(self):
        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mim://inmemdb'
        conf['ming.connection.read_preference'] = 'PRIMARY'

        app = conf.make_wsgi_app()
        assert app is not None

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_replica_set(self):
        if sys.version_info[:2] == (2, 6):
            pytest.skip()

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mongodb://localhost:27017,localhost:27018/testdb?replicaSet=test'
        conf['ming.db'] = ''

        app = conf.make_wsgi_app()
        assert app is not None

        expected_url = 'mongodb://localhost:27017,localhost:27018/testdb?replicaSet=test'
        expected_db = 'testdb'

        dstore = config['tg.app_globals'].ming_datastore
        assert expected_db == dstore.name, dstore.name
        assert dstore.bind._conn_args[0] == expected_url

    @pytest.mark.skipif(ming is None, reason="Ming not supported on this system")
    def test_setup_ming_persistance_replica_set_option(self):
        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_ming = True
        conf['ming.url'] = 'mongodb://localhost:27017,localhost:27018/testdb'
        conf['ming.connection.replicaSet'] = 'test'

        app = conf.make_wsgi_app()
        assert app is not None

        expected_url = 'mongodb://localhost:27017,localhost:27018/testdb'
        expected_db = 'testdb'

        dstore = config['tg.app_globals'].ming_datastore
        assert expected_db == dstore.name, dstore.name
        assert dstore.bind._conn_args[0] == expected_url
        assert 'test' == dstore.bind._conn_kwargs.get('replicaSet'), dstore.bind._conn_kwargs

    def test_setup_authtkt(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return str(request.environ)

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf["sa_auth.enabled"] = True
        conf['sa_auth'] = {'authmetadata': ApplicationAuthMetadataWithAuthentication(),
                           'dbsession': None,
                           'user_class': None,
                           'cookie_secret': '12345',
                           'post_login_url': '/'}
        conf['sqlalchemy.url'] = 'sqlite://'

        secure_app = conf.make_wsgi_app(**{'sa_auth.authtkt.secure': True})
        secure_app = TestApp(secure_app)
        resp = secure_app.post('/login_handler', params={'login': 'l', 'password': 'p'})
        assert 'HttpOnly' in resp.headers["Set-Cookie"], resp.headers

        insecure_app = conf.make_wsgi_app(**{'sa_auth.authtkt.secure': False})
        insecure_app = TestApp(insecure_app)
        resp = insecure_app.post('/login_handler', params={'login': 'l', 'password': 'p'})
        assert 'HttpOnly' not in resp.headers["Set-Cookie"], resp.headers


    def test_sessions_enabled(self):
        class RootController(TGController):
            @expose('json')
            def test(self):
                try:
                    tg.session['counter'] += 1
                except KeyError:
                    tg.session['counter'] = 0

                tg.session.save()
                return dict(counter=tg.session['counter'])

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['session.enabled'] = True
        app = conf.make_wsgi_app()
        app = TestApp(app)

        resp = app.get('/test')
        assert resp.json['counter'] == 0, resp

        resp = app.get('/test')
        assert resp.json['counter'] == 1, resp

    def test_caching_enabled(self):
        class RootController(TGController):
            @expose('json')
            def test(self):
                cache = tg.cache.get_cache('test_caching_enabled')
                now = cache.get_value('test_cache_key', createfunc=utcnow)
                return dict(now=now)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['cache.enabled'] = True
        app = conf.make_wsgi_app()
        app = TestApp(app)

        resp = app.get('/test')
        now = resp.json['now']

        for x in range(20):
            resp = app.get('/test')
            assert resp.json['now'] == now, (resp, now)

    def test_controler_wrapper_setup(self):
        from tg.configurator.components.dispatch import _call_controller
        orig_caller = _call_controller

        appcfg = AppConfig(minimal=True, root_controller=RootController())

        conf = {}
        dispatch = appcfg._configurator.get_component('dispatch')
        dispatch._controller_wrappers[:] = []
        dispatch._setup_controller_wrappers(conf, None)
        assert conf['controller_caller'] == orig_caller

        def controller_wrapper(caller):
            def call(*args, **kw):
                return caller(*args, **kw)
            return call

        conf = {}
        dispatch = appcfg._configurator.get_component('dispatch')
        dispatch._controller_wrappers[:] = [controller_wrapper]
        dispatch._setup_controller_wrappers(conf, None)
        assert conf['controller_caller'].__name__ == controller_wrapper(orig_caller).__name__

    def test_global_controller_wrapper(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(controller_wrapper)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] is True

    def test_multiple_global_controller_wrapper(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        def controller_wrapper2(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        def controller_wrapper3(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(controller_wrapper2)
        conf.register_controller_wrapper(controller_wrapper3)
        conf.register_controller_wrapper(controller_wrapper)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert len(wrapper_has_been_visited) == 3

    def test_dedicated_controller_wrapper(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(controller_wrapper, controller=RootController.test)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] is True

    def test_dedicated_controller_wrapper_old(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(controller_wrapper, controller=RootController.test)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] is True

    def test_mixed_controller_wrapper(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        app_wrapper_has_been_visited = []
        def app_controller_wrapper(caller):
            def call(*args, **kw):
                app_wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(app_controller_wrapper)
        conf.register_controller_wrapper(controller_wrapper, controller=RootController.test)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] is True
        assert app_wrapper_has_been_visited[0] is True

    def test_controler_wrapper_after_environment_setup(self):
        milestones._reset_all()

        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        def controller_wrapper(caller):
            def call(*args, **kw):
                wrapper_has_been_visited.append(True)
                return caller(*args, **kw)
            return call

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_controller_wrapper(controller_wrapper)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] is True
        assert len(wrapper_has_been_visited) == 1

        conf.register_controller_wrapper(controller_wrapper)
        app2 = conf.make_wsgi_app()
        app2 = TestApp(app2)

        wrapper_has_been_visited[:] = []
        assert 'HI!' in app2.get('/test')
        assert wrapper_has_been_visited[0] is True
        assert len(wrapper_has_been_visited) == 2

    def test_application_wrapper_setup(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        wrapper_has_been_visited = []
        class AppWrapper(object):
            def __init__(self, dispatcher):
                self.dispatcher = dispatcher
            def __call__(self, *args, **kw):
                wrapper_has_been_visited.append(True)
                return self.dispatcher(*args, **kw)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_wrapper(AppWrapper)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert wrapper_has_been_visited[0] == True

    def test_application_wrapper_ordering_after(self):
        class AppWrapper1:
            pass
        class AppWrapper2:
            pass
        class AppWrapper3:
            pass
        class AppWrapper4:
            pass
        class AppWrapper5:
            pass

        conf = AppConfig(minimal=True)
        conf.register_wrapper(AppWrapper2)
        conf.register_wrapper(AppWrapper4, after=AppWrapper3)
        conf.register_wrapper(AppWrapper3)
        conf.register_wrapper(AppWrapper1, after=False)
        conf.register_wrapper(AppWrapper5, after=AppWrapper3)
        milestones.environment_loaded.reach()

        app_wrappers = list(conf._configurator._application_wrappers.values())
        assert app_wrappers[0] == AppWrapper1
        assert app_wrappers[1] == AppWrapper2
        assert app_wrappers[2] == AppWrapper3
        assert app_wrappers[3] == AppWrapper4
        assert app_wrappers[4] == AppWrapper5

    def test_wrap_app(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        middleware_has_been_visited = []
        class AppWrapper(object):
            def __init__(self, app):
                self.app = app
            def __call__(self, environ, start_response):
                middleware_has_been_visited.append(True)
                return self.app(environ, start_response)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app(wrap_app=AppWrapper)
        app = TestApp(app)

        assert 'HI!' in app.get('/test')
        assert middleware_has_been_visited[0] == True

    def test_unsupported_renderer(self):
        conf = AppConfig(root_controller=RootController())
        conf['renderers'] = ['unknwon']

        try:
            conf.make_wsgi_app()
        except TGConfigError as e:
            assert 'This configuration object does not support the unknwon renderer' in str(e)
        else:
            assert False

    def test_cookie_secret_required(self):
        conf = AppConfig(root_controller=RootController())
        conf['sa_auth.enabled'] = True
        conf['sa_auth'] = {}
        try:
            conf.make_wsgi_app()
        except TGConfigError as e:
            assert str(e).startswith('You must provide a value for authentication cookies secret')
        else:
            assert False

    def test_sqla_auth_middleware_only_mine(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return str(request.environ)

            @expose()
            def forbidden(self):
                response.status = "401"

        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf['sqlalchemy.url'] = 'sqlite://'

        conf["sa_auth.enabled"] = True
        alwaysadmin = _AuthenticationForgerPlugin(fake_user_key='FAKE_USER')
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadata(),
                           'cookie_secret':'12345',
                           'form_plugin':alwaysadmin,
                           'authenticators':UncopiableList([('alwaysadmin', alwaysadmin)]),
                           'identifiers':[('alwaysadmin', alwaysadmin)],
                           'challengers':[]})

        app = conf.make_wsgi_app()

        authenticators = [x[0] for x in config['sa_auth.authenticators']]
        assert authenticators[0] == 'alwaysadmin'
        assert 'sqlauth' not in authenticators

        challengers = [x[1] for x in config['sa_auth.challengers']]
        assert alwaysadmin in challengers

        app = TestApp(app)
        assert 'repoze.who.identity' in app.get('/test', extra_environ={'FAKE_USER':'admin'})
        assert app.get('/forbidden', status=401)

    def test_sqla_auth_logging_stderr(self):
        package = PackageWithModel()
        conf = AppConfig(minimal=True, root_controller=None)
        conf.package = package
        conf.model = package.model
        conf.use_sqlalchemy = True
        conf['sqlalchemy.url'] = 'sqlite://'

        conf["sa_auth.enabled"] = True
        alwaysadmin = _AuthenticationForgerPlugin(fake_user_key='FAKE_USER')
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadata(),
                                'cookie_secret':'12345',
                                'form_plugin':alwaysadmin,
                                'log_level':'DEBUG',
                                'authenticators':UncopiableList([('alwaysadmin', alwaysadmin)]),
                                'identifiers':[('alwaysadmin', alwaysadmin)],
                                'challengers':[]})

        conf['sa_auth']['log_file'] = 'stderr'
        app = conf.make_wsgi_app()
        conf['sa_auth']['log_file'] = 'stdout'
        app = conf.make_wsgi_app()

        import tempfile
        f = tempfile.NamedTemporaryFile()
        conf['sa_auth']['log_file'] = f.name
        try:
            app = conf.make_wsgi_app()
        except OSError:
            # Ignore this error, as github actions don't allow to write temporary files.
            pytest.skip("GitHub doesn't allow to write temporary files")

    def test_sqla_auth_middleware_no_backend(self):
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = False
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadata(),
                                'cookie_secret':'12345'})
        conf.make_wsgi_app()

        with pytest.raises(KeyError):
            authenticators = [x[0] for x in config['sa_auth.authenticators']]
        #assert 'cookie' in authenticators
        #assert len(authenticators) == 1


    def test_sqla_auth_no_authenticate_meth(self):
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = True
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadata(),
                                'cookie_secret':'12345'})

        with pytest.raises(TGConfigError) as e:
            conf.make_wsgi_app()
        assert "missing authenticate" in str(e.value)

    def test_tgauthmetadata_auth_middleware(self):
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = True
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadataWithAuthentication(),
                                  'dbsession': None,
                                  'user_class':None,
                                  'cookie_secret':'12345',
                                  'authenticators':UncopiableList([('default', None)])})
        conf.make_wsgi_app()

        authenticators = [x[0] for x in config['sa_auth.authenticators']]
        assert 'cookie' in authenticators
        assert 'tgappauth' in authenticators

    def test_auth_setup_default_identifier(self):
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = True
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadataWithAuthentication(),
                                  'dbsession': None,
                                  'user_class':None,
                                  'cookie_secret':'12345',
                                  'identifiers': UncopiableList([('default', None)])})
        conf.make_wsgi_app()

        identifiers = [x[0] for x in tg.config['sa_auth.identifiers']]
        assert 'cookie' in identifiers

    def test_auth_setup_custom_identifier(self):
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = True
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadataWithAuthentication(),
                                'dbsession': None,
                                'user_class':None,
                                'cookie_secret':'12345',
                                'identifiers': UncopiableList([('custom', None)])})
        conf.make_wsgi_app()

        identifiers = [x[0] for x in config['sa_auth.identifiers']]
        assert 'custom' in identifiers

    def test_auth_middleware_doesnt_touch_authenticators(self):
        # Checks that the auth middleware process doesn't touch original authenticators
        # list, to prevent regressions on this.
        conf = AppConfig(root_controller=RootController())
        conf["sa_auth.enabled"] = True
        conf['sa_auth'].update({'authmetadata': ApplicationAuthMetadataWithAuthentication(),
                                'dbsession': None,
                                'user_class':None,
                                'cookie_secret':'12345',
                                'authenticators':[('default', None)]})
        conf.make_wsgi_app()

        authenticators = [x[0] for x in conf['sa_auth.authenticators']]
        assert len(authenticators) == 1

    def test_tgauthmetadata_loginpwd(self):
        who_authenticator = _AuthMetadataAuthenticator(ApplicationAuthMetadataWithAuthentication(), using_password=True)
        assert who_authenticator.authenticate({}, {}) == None

    def test_tgauthmetadata_nologinpwd(self):
        who_authenticator = _AuthMetadataAuthenticator(ApplicationAuthMetadataWithAuthentication(), using_password=False)
        assert who_authenticator.authenticate({}, {}) == 1

    def test_error_middleware_disabled_with_optimize(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()

        os.environ['PYTHONOPTIMIZE'] = '2'
        app = conf.make_wsgi_app()
        os.environ.pop('PYTHONOPTIMIZE')

        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_serve_statics(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        conf.serve_static = True
        app = conf.make_wsgi_app()
        assert app.__class__.__name__.startswith('Statics')

        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_mount_point_with_minimal(self):
        class SubController(TGController):
            @expose()
            def test(self):
                return self.mount_point

        class RootController(TGController):
            sub = SubController()

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()

        app = TestApp(app)
        assert '/sub' in app.get('/sub/test')

    def test_application_test_vars(self):
        class RootController(TGController):
            pass

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        assert 'DONE' in app.get('/_test_vars')
        assert request.path == '/_test_vars'

        # This should trash away the preserved registry to avoid
        # leaking memory.
        app.get('/', status=404)

        try:
            request.path
        except TypeError:
            # TypeError means the request has been properly removed
            pass
        else:
            assert False, 'There should have been no requests in place...'

    def test_application_empty_controller(self):
        class RootController(object):
            def __call__(self, environ, start_response):
                return None

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        try:
            r = app.get('/something')
        except Exception as e:
            assert 'No content returned by controller' in str(e)
        else:
            assert False, 'Should have raised "No content returned by controller"'

    def test_application_test_mode_detection(self):
        class FakeRegistry(object):
            def register(self, *args, **kw):
                pass

        def track_app(app):
            # Save a reference to the plain TGApp before it's wrapped by middlewares.
            track_app.app = app
            return app

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = PackageWithModel()
        conf.make_wsgi_app(wrap_app=track_app)

        testmode, __, __ = track_app.app._setup_app_env({'paste.registry':FakeRegistry()})
        assert testmode is False

        testmode, __, __ = track_app.app._setup_app_env({'paste.registry':FakeRegistry(),
                                                         'paste.testing_variables':{}})
        assert testmode is True

    def test_application_no_controller_hijacking(self):
        class RootController(TGController):
            @expose()
            def test(self):
                return 'HI!'

        class AppWrapper(object):
            def __init__(self, dispatcher):
                self.dispatcher = dispatcher
            def __call__(self, controller, environ, start_response):
                return self.dispatcher(None, environ, start_response)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_wrapper(AppWrapper)
        conf.package = PackageWithModel()
        app = conf.make_wsgi_app()
        app = TestApp(app)

        app.get('/test', status=404)

    def test_package_no_app_globals(self):
        class RootController(TGController):
            pass

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.package = sys.modules[__name__]

        app = conf.make_wsgi_app()

    def test_custom_error_document(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(403)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' in resp, resp

    def test_custom_error_document_with_streamed_response(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                response.status_code = 403
                def _output():
                    yield 'Hi'
                    yield 'World'
                return _output()

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' in resp, resp

    def test_error_document_passthrough(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                request.disable_error_pages()
                abort(403, detail='Custom Detail')

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'Custom Detail' in resp, resp

    def test_custom_old_error_document(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(403)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf.status_code_redirect = True
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' in resp, resp

    def test_custom_old_error_document_with_streamed_response(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                response.status_code = 403
                def _output():
                    yield 'Hi'
                    yield 'World'
                return _output()

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf.status_code_redirect = True
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' in resp, resp

    def test_custom_500_document(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(500)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['debug'] = False
        conf['errorpage.handle_exceptions'] = False
        conf['errorpage.status_codes'] += [500]
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500)
        assert 'ERROR!!!' in resp, resp

    def test_custom_500_document_on_crash(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                raise Exception('Crash!')

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['debug'] = False
        conf['errorpage.handle_exceptions'] = True
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500)
        assert 'ERROR!!!' in resp, resp

    def test_errorpage_reraises_exceptions(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                raise Exception('Crash!')

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['debug'] = False
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=False)
        app = TestApp(app)

        try:
            resp = app.get('/test', status=500)
        except Exception as e:
            assert 'Crash!' in str(e)
        else:
            assert False, 'Should have raised Crash! exception'

    def test_old_custom_500_document(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(500)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['debug'] = False
        conf.status_code_redirect = True
        conf['errorpage.enabled'] = True
        conf['errorpage.status_codes'] += [500]
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500)
        assert 'ERROR!!!' in resp, resp

    def test_skips_custom_500_document_when_debug(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(500)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['debug'] = True
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500)
        assert 'ERROR!!!' not in resp, resp

    def test_skips_old_custom_500_document_when_debug(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(500)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['debug'] = True
        conf.status_code_redirect = True
        conf['errorpage.enabled'] = True
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500)
        assert 'ERROR!!!' not in resp, resp

    def test_skips_custom_error_document_when_disabled(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(403)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = False
        conf['errorpage.status_codes'] = (403, 404)
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' not in resp, resp

    def test_skips_custom_error_document_when_disabled_and_manually_registered(self):
        class ErrorController(TGController):
            @expose()
            def document(self, *args, **kw):
                return 'ERROR!!!'

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(403)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = False
        conf['errorpage.status_codes'] = (403, 404)
        conf['errorpage.handle_exceptions'] = False
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=403)
        assert 'ERROR!!!' not in resp, resp

    def test_custom_500_json(self):
        class ErrorController(TGController):
            @expose(content_type="text/html")
            @expose('json', content_type="application/json")
            def document(self, *args, **kw):
                return dict(a=5)

        class RootController(TGController):
            error = ErrorController()
            @expose()
            def test(self):
                abort(500)

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['errorpage.enabled'] = True
        conf['debug'] = False
        conf['errorpage.handle_exceptions'] = False
        conf['errorpage.status_codes'] += [500]
        app = conf.make_wsgi_app(full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500,
                       headers={'Accept': 'application/json'})
        assert '{"a": 5}' in resp.text, resp
        assert 'application/json' == resp.content_type

    def test_errorware_configuration(self):
        class RootController(TGController):
            @expose()
            def test(self, *args, **kwargs):
                return 'HI'

        conf = AppConfig(minimal=True, root_controller=RootController())
        app = conf.make_wsgi_app(full_stack=True,
                                 **{'trace_errors.error_email': 'test@domain.com'})
        app = TestApp(app)
        resp = app.get('/test')
        assert 'HI' in resp, resp

        assert config['tg.errorware']['error_email'] == 'test@domain.com'
        assert config['tg.errorware']['error_subject_prefix'] == 'WebApp Error: '
        assert config['tg.errorware']['error_message'] == 'An internal server error occurred'

    def test_render_factory_success(self):
        class RootController(TGController):
            @expose()
            def test(self, *args, **kwargs):
                return 'HELLO'

        class FailedFactory(RendererFactory):
            engines = {'broken': {'content_type': 'text/plain'}}

            @classmethod
            def create(cls, config, app_globals):
                return {'broken': 'BROKEN'}

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_rendering_engine(FailedFactory)
        conf.renderers = ['json', 'broken']

        app = conf.make_wsgi_app(full_stack=True)
        assert config['renderers'] == ['json', 'broken']
        assert config['render_functions']['broken'] == 'BROKEN'

    def test_render_factory_failure(self):
        class RootController(TGController):
            @expose()
            def test(self, *args, **kwargs):
                return 'HELLO'

        class FailedFactory(RendererFactory):
            engines = {'broken': {'content_type': 'text/plain'}}

            @classmethod
            def create(cls, config, app_globals):
                return None

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.register_rendering_engine(FailedFactory)
        conf.renderers = ['json', 'broken']

        conf.make_wsgi_app(full_stack=True)
        assert config['renderers'] == ['json']

    def test_make_body_seekable(self):
        class RootController(TGController):
            @expose()
            def test(self, *args, **kwargs):
                request.body_file.seek(0)
                return 'HELLO'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['make_body_seekable'] = True

        app = conf.make_wsgi_app(full_stack=False)
        assert app.application.__class__.__name__ == 'SeekableRequestBodyMiddleware', \
            app.application.__class__

        app = TestApp(app)
        assert 'HELLO' in app.get('/test')

    def test_make_body_seekable_disabled(self):
        class RootController(TGController):
            @expose()
            def test(self, *args, **kwargs):
                request.body_file.seek(0)
                return 'HELLO'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf['make_body_seekable'] = False

        app = conf.make_wsgi_app(full_stack=False)
        app = TestApp(app)
        assert 'HELLO' in app.get('/test')

    def test_debug_middleware(self):
        class RootController(TGController):
            @expose()
            def test(self):
                raise Exception('Crash!')

        conf = AppConfig(root_controller=RootController())
        conf['errorpage.enabled'] = True
        app = conf.make_wsgi_app(debug=True, full_stack=True)
        app = TestApp(app)

        resp = app.get('/test', status=500, expect_errors=True)
        assert 'Exception: Crash! // Backlash' in resp, resp

    def test_make_app_with_custom_appglobals(self):
        class RootController(TGController):
            @expose('')
            def test(self, *args, **kwargs):
                return tg.app_globals.TEXT

        class FakeGlobals(Bunch):
            def __init__(self):
                super(FakeGlobals, self).__init__()
                self['TEXT'] = 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.app_globals = FakeGlobals
        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_make_app_with_appglobals_submodule(self):
        class RootController(TGController):
            @expose('')
            def test(self, *args, **kwargs):
                return tg.app_globals.text

        conf = AppConfig(minimal=True, root_controller=RootController())

        from .fixtures import package_with_helpers_submodule
        conf['package'] = package_with_helpers_submodule
        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!!' in app.get('/test')

    def test_make_app_with_custom_helpers(self):
        class RootController(TGController):
            @expose('')
            def test(self, *args, **kwargs):
                return config['helpers'].get_text()

        class FakeHelpers(object):
            @classmethod
            def get_text(cls):
                return 'HI!'

        conf = AppConfig(minimal=True, root_controller=RootController())
        conf.helpers = FakeHelpers()
        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!' in app.get('/test')

    def test_make_app_with_helpers_submodule(self):
        class RootController(TGController):
            @expose('')
            def test(self, *args, **kwargs):
                return config['helpers'].get_text()

        conf = AppConfig(minimal=True, root_controller=RootController())

        from .fixtures import package_with_helpers_submodule
        conf['package'] = package_with_helpers_submodule
        app = conf.make_wsgi_app()
        app = TestApp(app)
        assert 'HI!!' in app.get('/test')



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from tg.support.converters import asbool

from ...configuration.utils import TGConfigError
from ..base import ConfigurationComponent


class TransactionManagerConfigurationComponent(ConfigurationComponent):
    """Support for transaction manager.

    The transaction manager will automatically handle transactions
    bound to it according to the request state.

    Whenever a request succeeds the transaction manger will
    commit all transactions, while in case of a failure it will
    rollback the transactions.

    Refer to :class:`.TransactionApplicationWrapper` for list of
    supported options.
    """

    id = "transaction_manager"

    def get_defaults(self):
        return {"tm.enabled": True}

    def get_coercion(self):
        return {"tm.enabled": asbool}

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError(
                "Transactions Support only works on an ApplicationConfigurator"
            )

        from ...appwrappers.transaction_manager import TransactionApplicationWrapper

        configurator.register_application_wrapper(
            TransactionApplicationWrapper, after=True
        )



# -*- coding: utf-8 -*-
"""Helper functions for controller operation.

URL definition and browser redirection are defined here.

"""

from ..configuration.utils import TGConfigError
from ..request_local import request as tg_request
from ..request_local import response as tg_response
from ..support.responses import abort, etag_cache, redirect, validation_errors_response
from ..support.url import lurl, url


def use_wsgi_app(wsgi_app):
    return tg_request.get_response(wsgi_app)


def auth_force_login(user_name):
    """Forces user login if authentication is enabled.

    As TurboGears identifies users by ``user_name`` the passed parameter should
    be anything your application declares being the ``user_name`` field in models.

    """
    req = tg_request._current_obj()
    resp = tg_response._current_obj()

    api = req.environ.get("repoze.who.api")
    if api:
        authentication_plugins = req.environ["repoze.who.plugins"]
        try:
            identifier = authentication_plugins["main_identifier"]
        except KeyError:
            raise TGConfigError('No repoze.who plugin registered as "main_identifier"')

        resp.headers.extend(
            api.remember({"repoze.who.userid": user_name, "identifier": identifier})
        )


def auth_force_logout():
    """Forces user logout if authentication is enabled."""
    req = tg_request._current_obj()
    resp = tg_response._current_obj()

    api = req.environ.get("repoze.who.api")
    if api:
        resp.headers.extend(api.forget())


__all__ = (
    "url",
    "lurl",
    "redirect",
    "etag_cache",
    "abort",
    "auth_force_logout",
    "auth_force_login",
    "validation_errors_response",
    "use_wsgi_app",
)



from tg.configuration.auth.metadata import TGAuthMetadata, create_default_authenticator
from tg.configuration.auth.setup import _AuthenticationForgerPlugin, setup_auth

__all__ = (
    "TGAuthMetadata",
    "create_default_authenticator",
    "_AuthenticationForgerPlugin",
    "setup_auth",
)



"""This module contains the RestController implementation.

Rest controller provides a RESTful dispatch mechanism, and
combines controller decoration for TG-Controller behavior.
"""

from crank.restdispatcher import RestDispatcher

from ..decorators import expose
from .decoratedcontroller import DecoratedController
from .dispatcher import CoreDispatcher


class RestController(DecoratedController, CoreDispatcher, RestDispatcher):
    """A Decorated Controller that dispatches in a RESTful Manner.

    This controller was designed to follow Representational State Transfer protocol, also known as REST.
    The goal of this controller method is to provide the developer a way to map
    RESTful URLS to controller methods directly, while still allowing Normal Object Dispatch to occur.

    Here is a brief rundown of the methods which are called on dispatch along with an example URL.

    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | Method          | Description                                                  | Example Method(s) / URL(s)                 |
    +=================+==============================================================+============================================+
    | get_one         | Display one record.                                          | GET /movies/1                              |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | get_all         | Display all records in a resource.                           | GET /movies/                               |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | get             | A combo of get_one and get_all.                              | GET /movies/                               |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | GET /movies/1                              |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | new             | Display a page to prompt the User for resource creation.     | GET /movies/new                            |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | edit            | Display a page to prompt the User for resource modification. |  GET /movies/1/edit                        |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | post            | Create a new record.                                         | POST /movies/                              |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | put             | Update an existing record.                                   | POST /movies/1?_method=PUT                 |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | PUT /movies/1                              |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | post_delete     | Delete an existing record.                                   | POST /movies/1?_method=DELETE              |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | DELETE /movies/1                           |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | get_delete      | Display a delete Confirmation page.                          | GET /movies/1/delete                       |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+
    | delete          | A combination of post_delete and get_delete.                 | GET /movies/delete                         |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | DELETE /movies/1                           |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | DELETE /movies/                            |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | POST /movies/1/delete                      |
    |                 |                                                              +--------------------------------------------+
    |                 |                                                              | POST /movies/delete                        |
    +-----------------+--------------------------------------------------------------+--------------------------------------------+

    You may note the ?_method on some of the URLs.  This is basically a hack because exiting browsers
    do not support the PUT and DELETE methods.  Just note that if you decide to use a this resource with a web browser,
    you will likely have to add a _method as a hidden field in your forms for these items.  Also note that RestController differs
    from TGController in that it offers no index, _default, or _lookup.  It is intended primarily for  resource management.

    :References:

      `Controller <../main/Controllers.html>`_  A basic overview on how to write controller methods.

      `CrudRestController <../main/Extensions/Crud/index.html>`_  A way to integrate ToscaWdiget Functionality with RESTful Dispatch.

    """

    @expose()
    def options(self, *args, **kw):
        """The OPTIONS http verb is automatically sent by the browser when doing cross origin requests.

        This default method was added to prevent receiving 404 HTTP Not found error.
        It returns None which translates to 204 No Content for HTTP.
        You can use the _before hook to set up your CORS headers.

        """
        return None


__all__ = ["RestController"]



# -*- coding: utf-8 -*-
"""
Built-in predicate checkers.

This is mostly took from repoze.what.precidates

This is module provides the predicate checkers that were present in the
original "identity" framework of TurboGears 1, plus others.

"""

from .request_local import request

__all__ = [
    "Predicate",
    "CompoundPredicate",
    "All",
    "Any",
    "has_all_permissions",
    "has_any_permission",
    "has_permission",
    "in_all_groups",
    "in_any_group",
    "in_group",
    "is_user",
    "is_anonymous",
    "not_anonymous",
    "NotAuthorizedError",
]


class NotAuthorizedError(Exception):
    pass


class Predicate(object):
    def __init__(self, msg=None):
        if msg:
            self.message = msg

    def evaluate(self, environ, credentials):
        raise NotImplementedError

    def unmet(self, msg=None, **placeholders):
        """
        Raise an exception because this predicate is not met.

        :param msg: The error message to be used; overrides the predicate's
            default one.
        :type msg: str
        :raises NotAuthorizedError: If the predicate is not met.

        ``placeholders`` represent the placeholders for the predicate message.
        The predicate's attributes will also be taken into account while
        creating the message with its placeholders.
        """
        if msg:
            message = msg
        else:
            message = self.message

        # This enforces lazy strings resolution (lazy translation for example)
        message = str(message)

        # Include the predicate attributes in the placeholders:
        all_placeholders = self.__dict__.copy()
        all_placeholders.update(placeholders)

        raise NotAuthorizedError(message % all_placeholders)

    def check_authorization(self, environ):
        """
        Evaluate the predicate and raise an exception if it's not met.

        :param environ: The WSGI environment.
        :raise NotAuthorizedError: If it the predicate is not met.
        """
        credentials = environ.get("repoze.what.credentials", {})
        try:
            self.evaluate(environ, credentials)
        except NotAuthorizedError:
            raise

    def is_met(self, environ):
        """
        Find whether the predicate is met or not.

        :param environ: The WSGI environment.
        :return: Whether the predicate is met or not.
        :rtype: bool
        """
        credentials = environ.get("repoze.what.credentials", {})
        try:
            self.evaluate(environ, credentials)
            return True
        except NotAuthorizedError:
            return False

    def __nonzero__(self):
        return self.is_met(request.environ)

    __bool__ = __nonzero__


class CompoundPredicate(Predicate):
    """A predicate composed of other predicates."""

    def __init__(self, *predicates, **kwargs):
        super(CompoundPredicate, self).__init__(**kwargs)
        self.predicates = predicates


class Not(Predicate):
    """
    Negate the specified predicate.

    :param predicate: The predicate to be negated.

    Example::

        # The user *must* be anonymous:
        p = Not(not_anonymous())

    """

    message = "The condition must not be met"

    def __init__(self, predicate, **kwargs):
        super(Not, self).__init__(**kwargs)
        self.predicate = predicate

    def evaluate(self, environ, credentials):
        try:
            self.predicate.evaluate(environ, credentials)
        except NotAuthorizedError:
            return
        self.unmet()


class All(CompoundPredicate):
    """
    Check that all of the specified predicates are met.

    :param predicates: All of the predicates that must be met.

    Example::

        # Grant access if the current month is July and the user belongs to
        # the human resources group.
        p = All(is_month(7), in_group('hr'))

    """

    def evaluate(self, environ, credentials):
        """
        Evaluate all the predicates it contains.

        :param environ: The WSGI environment.
        :param credentials: The :mod:`repoze.what` ``credentials``.
        :raises NotAuthorizedError: If one of the predicates is not met.

        """
        for p in self.predicates:
            p.evaluate(environ, credentials)


class Any(CompoundPredicate):
    """
    Check that at least one of the specified predicates is met.

    :param predicates: Any of the predicates that must be met.

    Example::

        # Grant access if the currest user is Richard Stallman or Linus
        # Torvalds.
        p = Any(is_user('rms'), is_user('linus'))

    """

    message = (
        "At least one of the following predicates must be met: %(failed_predicates)s"
    )

    def evaluate(self, environ, credentials):
        """
        Evaluate all the predicates it contains.

        :param environ: The WSGI environment.
        :param credentials: The :mod:`repoze.what` ``credentials``.
        :raises NotAuthorizedError: If none of the predicates is met.

        """
        errors = []
        for p in self.predicates:
            try:
                p.evaluate(environ, credentials)
                return
            except NotAuthorizedError as exc:
                errors.append(str(exc))
        failed_predicates = ", ".join(errors)
        self.unmet(failed_predicates=failed_predicates)


class is_user(Predicate):
    """
    Check that the authenticated user's username is the specified one.

    :param user_name: The required user name.
    :type user_name: str

    Example::

        p = is_user('linus')

    """

    message = 'The current user must be "%(user_name)s"'

    def __init__(self, user_name, **kwargs):
        super(is_user, self).__init__(**kwargs)
        self.user_name = user_name

    def evaluate(self, environ, credentials):
        if credentials and self.user_name == credentials.get("repoze.what.userid"):
            return
        self.unmet()


class in_group(Predicate):
    """
    Check that the user belongs to the specified group.

    :param group_name: The name of the group to which the user must belong.
    :type group_name: str

    Example::

        p = in_group('customers')

    """

    message = 'The current user must belong to the group "%(group_name)s"'

    def __init__(self, group_name, **kwargs):
        super(in_group, self).__init__(**kwargs)
        self.group_name = group_name

    def evaluate(self, environ, credentials):
        if credentials and self.group_name in credentials.get("groups"):
            return
        self.unmet()


class in_all_groups(All):
    """
    Check that the user belongs to all of the specified groups.

    :param groups: The name of all the groups the user must belong to.

    Example::

        p = in_all_groups('developers', 'designers')

    """

    def __init__(self, *groups, **kwargs):
        group_predicates = [in_group(g) for g in groups]
        super(in_all_groups, self).__init__(*group_predicates, **kwargs)


class in_any_group(Any):
    """
    Check that the user belongs to at least one of the specified groups.

    :param groups: The name of any of the groups the user may belong to.

    Example::

        p = in_any_group('directors', 'hr')

    """

    message = (
        "The member must belong to at least one of the following groups: %(group_list)s"
    )

    def __init__(self, *groups, **kwargs):
        self.group_list = ", ".join(groups)
        group_predicates = [in_group(g) for g in groups]
        super(in_any_group, self).__init__(*group_predicates, **kwargs)


class is_anonymous(Predicate):
    """
    Check that the current user is anonymous.

    Example::

        # The user must be anonymous!
        p = is_anonymous()

    .. versionadded:: 1.0.7

    """

    message = "The current user must be anonymous"

    def evaluate(self, environ, credentials):
        if credentials:
            self.unmet()


class not_anonymous(Predicate):
    """
    Check that the current user has been authenticated.

    Example::

        # The user must have been authenticated!
        p = not_anonymous()

    """

    message = "The current user must have been authenticated"

    def evaluate(self, environ, credentials):
        if not credentials:
            self.unmet()


class has_permission(Predicate):
    """
    Check that the current user has the specified permission.

    :param permission_name: The name of the permission that must be granted to
        the user.

    Example::

        p = has_permission('hire')

    """

    message = 'The user must have the "%(permission_name)s" permission'

    def __init__(self, permission_name, **kwargs):
        super(has_permission, self).__init__(**kwargs)
        self.permission_name = permission_name

    def evaluate(self, environ, credentials):
        if credentials and self.permission_name in credentials.get("permissions"):
            return
        self.unmet()


class has_all_permissions(All):
    """
    Check that the current user has been granted all of the specified
    permissions.

    :param permissions: The names of all the permissions that must be
        granted to the user.

    Example::

        p = has_all_permissions('view-users', 'edit-users')

    """

    def __init__(self, *permissions, **kwargs):
        permission_predicates = [has_permission(p) for p in permissions]
        super(has_all_permissions, self).__init__(*permission_predicates, **kwargs)


class has_any_permission(Any):
    """
    Check that the user has at least one of the specified permissions.

    :param permissions: The names of any of the permissions that have to be
        granted to the user.

    Example::

        p = has_any_permission('manage-users', 'edit-users')

    """

    message = "The user must have at least one of the following permissions: %(permission_list)s"

    def __init__(self, *permissions, **kwargs):
        self.permission_list = ", ".join(permissions)
        permission_predicates = [has_permission(p) for p in permissions]
        super(has_any_permission, self).__init__(*permission_predicates, **kwargs)



from __future__ import absolute_import

from markupsafe import Markup

import tg
from tg.configuration.utils import coerce_config
from tg.i18n import ugettext
from tg.render import cached_template
from tg.support.converters import asbool, asint

from .base import RendererFactory

try:
    import genshi
except ImportError:  # pragma: no cover
    genshi = None

if genshi is not None:
    from genshi import HTML, XML
    from genshi.filters import Translator
    from genshi.template import TemplateLoader as GenshiTemplateLoader
else:  # pragma: no cover

    class GenshiTemplateLoader(object):
        pass


__all__ = ["GenshiRenderer"]


class GenshiRenderer(RendererFactory):
    """
    Configuration Options available as ``templating.genshi.*``:

        - ``templating.genshi.name_constant_patch`` -> Enable/Disable patch for Python3.4 compatibility.
        - ``templating.genshi.max_cache_size`` -> Maximum number of templates to keep cached, by default 30.
        - ``templating.genshi.method`` -> Genshi rendering method (html or xhtml).

    Supported ``render_params``:

        - Caching options supported by :func:`.cached_template`
        - ``doctype`` -> To override the global doctype
        - ``method`` -> To override the global rendering method
    """

    CONFIG_OPTIONS = {"max_cache_size": asint, "name_constant_patch": asbool}
    engines = {"genshi": {"content_type": "text/html"}}

    doctypes_for_methods = {"html": "html-transitional", "xhtml": "xhtml-transitional"}

    doctypes_for_content_type = {
        "text/html": (
            "html",
            "html-transitional",
            "html-frameset",
            "html5",
            "xhtml",
            "xhtml-strict",
            "xhtml-transitional",
            "xhtml-frameset",
        ),
        "application/xhtml+xml": (
            "xhtml",
            "xhtml-strict",
            "xhtml-transitional",
            "xhtml-frameset",
            "xhtml11",
        ),
        "image/svg+xml": ("svg", "svg-full", "svg-basic", "svg-tiny"),
    }

    methods_for_content_type = {
        "text/plain": ("text",),
        "text/css": ("text",),
        "text/html": ("html", "xhtml"),
        "text/xml": ("xml", "xhtml"),
        "application/xml": ("xml", "xhtml"),
        "application/xhtml+xml": ("xhtml",),
        "application/atom+xml": ("xml",),
        "application/rss+xml": ("xml",),
        "application/soap+xml": ("xml",),
        "image/svg+xml": ("xml",),
    }

    @classmethod
    def create(cls, config, app_globals):
        """Setup a renderer and loader for Genshi templates.

        Override this to customize the way that the internationalization
        filter, template loader

        """
        if genshi is None:  # pragma: no cover
            # Genshi not available
            return None

        options = coerce_config(config, "templating.genshi.", cls.CONFIG_OPTIONS)

        # Patch for Genshi on Python3.4
        if options.get("name_constant_patch", False):  # pragma: no cover
            from genshi.template.astutil import ASTCodeGenerator

            if not hasattr(ASTCodeGenerator, "visit_NameConstant"):

                def _visit_NameConstant(self, node):
                    if node.value is None:
                        self._write("None")
                    elif node.value is True:
                        self._write("True")
                    elif node.value is False:
                        self._write("False")
                    else:
                        raise Exception("Unknown NameConstant %r" % (node.value,))

                ASTCodeGenerator.visit_NameConstant = _visit_NameConstant

        if config.get("use_dotted_templatenames", True):
            TemplateLoader = DottedTemplateLoader
            template_loader_args = {"dotted_finder": app_globals.dotted_filename_finder}
        else:
            TemplateLoader = GenshiTemplateLoader
            template_loader_args = {}

        loader = TemplateLoader(
            search_path=config["paths"].templates,
            max_cache_size=options.get(
                "max_cache_size", asint(config.get("genshi.max_cache_size", 30))
            ),
            auto_reload=config["auto_reload_templates"],
            callback=cls.on_template_loaded,
            **template_loader_args,
        )

        return {"genshi": cls(loader, config)}

    def __init__(self, loader, config):
        self.tg_config = config
        self.genshi_functions = dict(HTML=HTML, XML=XML)
        self.load_template = loader.load

        self.default_doctype = None
        doctype = self.tg_config.get("templating.genshi.doctype")
        if doctype:
            if isinstance(doctype, str):
                self.default_doctype = doctype
            elif isinstance(doctype, dict):
                doctypes = self.doctypes_for_content_type.copy()
                doctypes.update(doctype)
                self.doctypes_for_content_type = doctypes

        self.default_method = None
        method = self.tg_config.get("templating.genshi.method")
        if method:
            if isinstance(method, str):
                self.default_method = method
            elif isinstance(method, dict):
                methods = self.methods_for_content_type.copy()
                methods.update(method)
                self.methods_for_content_type = methods

    @classmethod
    def on_template_loaded(cls, template):
        """
        Plug-in our i18n function to Genshi, once the template is loaded.

        This function will be called by the Genshi TemplateLoader after
        loading the template.

        """
        translator = Translator(ugettext)
        template.filters.insert(0, translator)

        if hasattr(template, "add_directives"):
            template.add_directives(Translator.NAMESPACE, translator)

    @staticmethod
    def method_for_doctype(doctype):
        method = "xhtml"
        if doctype:
            if doctype.startswith("html"):
                method = "html"
            elif doctype.startswith("xhtml"):
                method = "xhtml"
            elif doctype.startswith("svg"):
                method = "xml"
            else:
                method = "xhtml"
        return method

    def __call__(self, template_name, template_vars, **kwargs):
        """Render the template_vars with the Genshi template.

        If you don't pass a doctype or pass 'auto' as the doctype,
        then the doctype will be automatically determined.
        If you pass a doctype of None, then no doctype will be injected.
        If you don't pass a method or pass 'auto' as the method,
        then the method will be automatically determined.

        """
        response = tg.response._current_obj()

        template_vars.update(self.genshi_functions)

        # Gets document type from content type or from config options
        doctype = kwargs.get("doctype", "auto")
        if doctype == "auto":
            doctype = self.default_doctype
            if not doctype:
                method = kwargs.get("method") or self.default_method or "xhtml"
                doctype = self.doctypes_for_methods.get(method)
            doctypes = self.doctypes_for_content_type.get(response.content_type)
            if doctypes and (not doctype or doctype not in doctypes):
                doctype = doctypes[0]
            kwargs["doctype"] = doctype

        # Gets rendering method from content type or from config options
        method = kwargs.get("method")
        if not method or method == "auto":
            method = self.default_method
            if not method:
                method = self.method_for_doctype(doctype)
            methods = self.methods_for_content_type.get(response.content_type)
            if methods and (not method or method not in methods):
                method = methods[0]
            kwargs["method"] = method

        def render_template():
            template = self.load_template(template_name)
            return Markup(
                template.generate(**template_vars).render(
                    doctype=doctype, method=method, encoding=None
                )
            )

        return cached_template(
            template_name, render_template, ns_options=("doctype", "method"), **kwargs
        )


class DottedTemplateLoader(GenshiTemplateLoader):
    """
    Genshi template loader supporting dotted filenames.
    Supports zipped applications and dotted filenames as well as path names.

    """

    def __init__(self, *args, **kwargs):
        self.template_extension = kwargs.pop("template_extension", ".html")
        self.dotted_finder = kwargs.pop("dotted_finder")

        super(DottedTemplateLoader, self).__init__(*args, **kwargs)

    def get_dotted_filename(self, filename):
        if not filename.endswith(self.template_extension):
            finder = self.dotted_finder
            filename = finder.get_dotted_filename(
                template_name=filename, template_extension=self.template_extension
            )
        return filename

    def load(self, filename, relative_to=None, cls=None, encoding=None):
        """Actual loader function."""
        return super(DottedTemplateLoader, self).load(
            self.get_dotted_filename(filename),
            relative_to=relative_to,
            cls=cls,
            encoding=encoding,
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import mimetypes

from ..base import BeforeConfigConfigurationAction, ConfigurationComponent


class MimeTypesConfigurationComponent(ConfigurationComponent):
    """Configure known MimeTypes.

    Mimetypes are used by turbogears to detect expected content
    types based on file extensions. For example it's used by
    :class:`.DispatchConfigurationComponent` request extensions
    to serve the right content based on URL path extension.

    Options:

        * ``mimetype_lookup``: Additional mapping from extensions to
                               mimetypes that should be configured.
    """

    id = "mimetypes"

    def get_actions(self):
        return (BeforeConfigConfigurationAction(self._configure_mimetypes),)

    def _configure_mimetypes(self, conf, app):
        conf["mimetypes"] = mimetypes.MimeTypes()

        lookup = {".json": "application/json", ".js": "application/javascript"}
        lookup.update(conf.get("mimetype_lookup", {}))

        for key, value in lookup.items():
            conf["mimetypes"].add_type(value, key)



# -*- coding: utf-8 -*-
from tg.support.converters import asbool

from ..base import AppReadyConfigurationAction, ConfigurationComponent


class SeekableRequestConfigurationComponent(ConfigurationComponent):
    """Support for making the request body seekable.

    This allows to make the request body seekable, so that it can be
    read multiple times and it's possible to go back and forth in
    request submitted data. Note that this has a cost in terms of
    consumed memory.

    Options:

        * ``make_body_seekable``: Enable seekable request body.
                                  By default this is disabled.

    """

    id = "seekable_request"

    def get_coercion(self):
        return {"make_body_seekable": asbool}

    def get_defaults(self):
        return {"make_body_seekable": False}

    def get_actions(self):
        return (AppReadyConfigurationAction(self._add_middleware),)

    def _add_middleware(self, conf, app):
        if conf["make_body_seekable"]:
            from tg.support.middlewares import SeekableRequestBodyMiddleware

            return SeekableRequestBodyMiddleware(app)
        return app



import logging

from ..configuration.utils import coerce_config
from ..support.converters import asbool
from .base import ApplicationWrapper

log = logging.getLogger(__name__)


class IdentityApplicationWrapper(ApplicationWrapper):
    """Provides user identity when authentication is enabled.

    The repoze.who provided identity takes precedence over the identity
    provided by IdentityApplicationWrapper if available.

    Supported options which can be provided by config are:
        - ``sa_auth.authmetadata``: The TGAuthMetadata object that should be used to retrieve identity metadata.
        - ``identity.enabled``: Enable the Identity Application Wrapper. By default enabled if authmetadata available.
        - ``identity.allow_missing_user``: Whenever the identity should be discarded or not when the authmetadata is unable to find an user.

    """

    def __init__(self, handler, config):
        super(IdentityApplicationWrapper, self).__init__(handler, config)

        options = {
            "enabled": True,
            "allow_missing_user": True,
            "authmetadata": config.get("sa_auth.authmetadata"),
        }
        options.update(
            coerce_config(
                config, "identity.", {"enabled": asbool, "allow_missing_user": asbool}
            )
        )

        self.enabled = options["enabled"] and options["authmetadata"] is not None
        self.options = options
        self.tgmdprovider = options["authmetadata"]
        log.debug("Identity enabled: %s -> %s", self.enabled, self.options)

    @property
    def injected(self):
        return self.enabled

    def __call__(self, controller, environ, context):
        identity = environ.get("repoze.who.identity")
        if identity is None:
            context.request.identity = None
            return self.next_handler(controller, environ, context)

        req_identity = {}

        # Get the userid retrieved by repoze.who Authenticator
        userid = identity["repoze.who.userid"]
        if userid is not None:
            # Finding the user, groups and permissions:
            identity["user"] = identity_user = self.tgmdprovider.get_user(
                identity, userid
            )

            if identity_user:
                identity["groups"] = self.tgmdprovider.get_groups(identity, userid)
                identity["permissions"] = self.tgmdprovider.get_permissions(
                    identity, userid
                )
            else:
                identity["groups"] = identity["permissions"] = []

            req_identity = Identity()
            req_identity.update(identity)
            req_identity["repoze.what.userid"] = userid

            if (
                req_identity.get("user") is None
                and not self.options["allow_missing_user"]
            ):
                req_identity = {}

        # Add identity to request with repoze.who/what compatibility
        context.request.identity = req_identity
        environ["repoze.who.identity"] = req_identity
        environ["repoze.what.credentials"] = req_identity

        return self.next_handler(controller, environ, context)


class Identity(dict):
    """dict subclass: prevent members from being rendered during print.

    Took as is from repoze.who.
    """

    def __repr__(self):
        return "<TurboGears Identity (hidden, dict-like) at %s>" % id(self)

    __str__ = __repr__



# -*- coding: utf-8 -*-
from importlib import import_module
from logging import getLogger

from ...util import Bunch
from ..base import ConfigReadyConfigurationAction, ConfigurationComponent

log = getLogger(__name__)


class HelpersConfigurationComponent(ConfigurationComponent):
    """Add support for helpers in templates.

    Helpers are a bunch of help functions, usually involved
    in generating small HTML snippets or formatting data
    that are exposed in templates under the ``h`` object.

    By default the helpers will be loaded from your application
    package as ``package.lib.helpers`` which is expected
    to be a module containing all helpers functions.

    Supported options:

        * ``helpers``: Set a different object as the template helpers.

    """

    id = "helpers"

    def get_defaults(self):
        return {
            "helpers": None,
        }

    def get_actions(self):
        return (ConfigReadyConfigurationAction(self._setup_helpers),)

    def _setup_helpers(self, conf, app):
        """Add helpers and globals objects to the ``conf``.

        Override this method to customize the way that ``app_globals`` and ``helpers``
        are setup. TurboGears expects them to be available in ``conf`` dictionary
        as ``tg.app_globals`` and ``helpers``.
        """
        # Setup Helpers
        h = conf.get("helpers", None)
        if h is None:
            try:
                h = conf["package"].lib.helpers
            except AttributeError:
                pass

        if h is None:
            try:
                h = import_module(".lib.helpers", package=conf["package"].__name__)
            except (ImportError, AttributeError):
                pass

        if h is None:
            log.warning("helpers not provided and lib.helpers is not available.")
            h = Bunch()

        conf["helpers"] = h



import inspect
import itertools
from collections import deque

from .milestones import config_ready


class TGConfigError(Exception):
    pass


def coerce_options(options, converters):
    """Convert some configuration options to expected types.

    To replace given options with the converted values
    in a dictionary you might do::

        conf.update(coerce_options(conf, {
            'debug': asbool,
            'serve_static': asbool,
            'auto_reload_templates': asbool
        }))
    """
    converted_options = {}
    for option, converter in converters.items():
        if option in options:
            converted_options[option] = converter(options[option])
    return converted_options


def coerce_config(configuration, prefix, converters):
    """Extracts a set of options with a common prefix and converts them.

    To extract all options starting with ``trace_errors.`` from
    the ``conf`` dictionary and conver them::

        trace_errors_config = coerce_config(conf, 'trace_errors.', {
            'smtp_use_tls': asbool,
            'dump_request_size': asint,
            'dump_request': asbool,
            'dump_local_frames': asbool,
            'dump_local_frames_count': asint
        })
    """

    options = dict(
        (key[len(prefix) :], configuration[key])
        for key in configuration
        if key.startswith(prefix)
    )
    options.update(coerce_options(options, converters))
    return options


class GlobalConfigurable(object):
    """Defines a configurable TurboGears object with a global default instance.

    GlobalConfigurable are objects which the user can create multiple instances to use
    in its own application or third party module, but for which TurboGears provides
    a default instance.

    Common examples are ``tg.flash`` and the default JSON encoder for which
    TurboGears provides default instances of ``.TGFlash`` and ``.JSONEncoder`` classes
    but users can create their own.

    While user created versions are configured calling the :meth:`.GlobalConfigurable.configure`
    method, global versions are configured by :class:`.ApplicationConfigurator` which
    configures them when ``config_ready`` milestone is reached.

    """

    CONFIG_NAMESPACE = None
    CONFIG_OPTIONS = {}

    def configure(self, **options):
        """Expected to be implemented by each object to proceed with actualy configuration.

        Configure method will receive all the options whose name starts with ``CONFIG_NAMESPACE``
        (example ``json.isodates`` has ``json.`` namespace).

        If ``CONFIG_OPTIONS`` is specified options values will be converted with
        :func:`coerce_config` passing ``CONFIG_OPTIONS`` as the ``converters`` dictionary.

        """
        raise NotImplementedError(
            "GlobalConfigurable objects must implement a configure method"
        )

    @classmethod
    def create_global(cls):
        """Creates a global instance whose configuration is linked to ``tg.config``."""
        if cls.CONFIG_NAMESPACE is None:
            raise TGConfigError(
                "Must specify a CONFIG_NAMESPACE attribute in class for the"
                "namespace used by all configuration options."
            )

        obj = cls()
        config_ready.register(obj._load_config, persist_on_reset=True)
        return obj

    def _load_config(self):
        from tg.configuration import config

        self.configure(
            **coerce_config(config, self.CONFIG_NAMESPACE, self.CONFIG_OPTIONS)
        )


class DependenciesList(object):
    """Manages a list of entries which might depend one from the other.

    This powers everything that needs to keep track of a list of entries that
    might have dependencies one to the other and keeps them linearised according
    to the dependencies. This powers the resolution of ConfigurationComponenets,
    ApplicationsWrappers and so on...

    .. note:: This is highly inefficient as it is only meant to run at configuration time,
              a new implementation will probably be provided based on heapq in the future.
    """

    #: Those are the heads of the dependencies tree
    #:  - ``False`` means before everything else
    #:  - ``None`` means in the middle.
    #:  - ``True`` means after everything else.
    DEPENDENCY_HEADS = (False, None, True)

    def __init__(self, *entries):
        self._dependencies = {}
        self._ordered_elements = []
        self._inserted_keys = []

        for entry in entries:
            self.add(entry)

    def add(self, entry, key=None, after=None):
        """Adds an entry to the dependencies list.

        :param entry: Entry that must be added to the list.
        :param str|type|None key: An identifier of the object being inserted.
                                  This is used by later insertions as ``after`` argument
                                  to specify after which object the new one should be inserted.
        :param str|type|None|False|True after: After which element this one should be inserted.
                                               This is the ``key`` of a previously inserted item.
                                               In case no item with ``key`` has been inserted, the
                                               entry will be inserted in normal order of insertion.
                                               Also accepts one of
                                               :attr:`.DependenciesList.DEPENDENCY_HEADS` as key
                                               to add entries at begin or end of the list.
        """
        if key is None:
            if inspect.isclass(entry):
                key = entry.__name__
            else:
                # Inserting an object without a key would lead to unexpected ordering.
                # we cannot use the object class as the key would not be unique across
                # different instances.
                raise ValueError("Inserting instances without a key is not allowed")

        if after not in self.DEPENDENCY_HEADS and not isinstance(after, str):
            if inspect.isclass(after):
                after = after.__name__
            else:
                raise ValueError(
                    "after parameter must be a string, a class or a special value"
                )

        if key in self._inserted_keys:
            raise KeyError("Already existing entry for this key")

        self._inserted_keys.append(key)
        self._dependencies.setdefault(after, []).append((key, entry))
        self._resolve_ordering()

    def __repr__(self):
        return "<DependenciesList %s>" % [x[0] for x in self._ordered_elements]

    def __iter__(self):
        return iter(self._ordered_elements)

    def values(self):
        """Returns all the inserted values without their key as a generator"""
        return (x[1] for x in self._ordered_elements)

    def replace(self, key, newvalue):
        """Replaces entry associated to key with a new one.

        :param newvalue: Entry that must replace the previous value.
        :param str|type key: An identifier of the object being inserted.
        """
        if not isinstance(key, str):
            if inspect.isclass(key):
                key = key.__name__
            else:
                raise ValueError("key parameter must be a string or a class")

        for entries in self._dependencies.values():
            for idx, value in enumerate(entries):
                entry_key, entry_value = value
                if entry_key == key:
                    entries[idx] = (entry_key, newvalue)

        self._resolve_ordering()

    def get(self, key):
        if not isinstance(key, str):
            if inspect.isclass(key):
                key = key.__name__
            else:
                raise ValueError("key parameter must be a string or a class")

        for entries in self._dependencies.values():
            for idx, value in enumerate(entries):
                entry_key, entry_value = value
                if entry_key == key:
                    return entry_value

        return None

    def _resolve_ordering(self):
        ordered_elements = []

        existing_dependencies = set(self._inserted_keys) | set(self.DEPENDENCY_HEADS)
        dependencies_without_heads = set(self._dependencies.keys()) - set(
            self.DEPENDENCY_HEADS
        )

        # All entries that depend on a missing entry are converted
        # to depend from None so they end up being in the middle.
        dependencies = {}
        for dependency in itertools.chain(
            self.DEPENDENCY_HEADS, dependencies_without_heads
        ):
            entries = self._dependencies.get(dependency, [])
            if dependency not in existing_dependencies:
                dependency = None
            dependencies.setdefault(dependency, []).extend(entries)

        # Resolve the dependencies and generate the ordered elements.
        visit_queue = deque((head, head) for head in self.DEPENDENCY_HEADS)
        while visit_queue:
            current_key, current_obj = visit_queue.popleft()
            if current_key not in self.DEPENDENCY_HEADS:
                ordered_elements.append((current_key, current_obj))

            element_dependencies = dependencies.pop(current_key, [])
            visit_queue.extendleft(reversed(element_dependencies))

        self._ordered_elements = ordered_elements


class DictionaryView(object):
    """Provides a view on a slice of a dictionary.

    This allows to expose all keys in a dictionary
    that start with a common prefix as a subdictionary.

    For example you might expose all keys starting with
    sqlalchemy.* as a standalone dictionary, all changes
    to the view will reflect into the dictionary updating
    the original keys.
    """

    __slots__ = ("_d", "_keypath")

    def __init__(self, d, keypath):
        if keypath and keypath[-1] != ".":
            keypath = keypath + "."
        self._d = d
        self._keypath = keypath

    def __getitem__(self, item):
        key = self._keypath + item
        return self._d[key]

    def __setitem__(self, key, value):
        key = self._keypath + key
        self._d[key] = value

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            key = self._keypath + item
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if key not in self.__slots__:
            self.__setitem__(key, value)
        else:
            object.__setattr__(self, key, value)

    def update(self, d, **d2):
        if hasattr(d, "keys"):
            for key in d.keys():
                self[key] = d[key]
        else:
            for key, value in d:
                self[key] = value

        for key in d2:
            self[key] = d2[key]


def copyoption(v):
    """Copies a dictionary and all its nested dictionaries and lists.

    Much like copy.deepcopy it provides a deep copy of a dictionary
    but instead of trying to copy everything it will only make a copy
    of dictionaries, lists, tuples and sets. All the containers typically
    used in configuration blueprints. All the other objects will be
    preserved by reference.
    """
    if isinstance(v, dict):
        return v.__class__((k, copyoption(v[k])) for k in v)
    elif isinstance(v, (list, set, tuple)):
        return v.__class__(copyoption(e) for e in v)
    else:
        # Preserve anything else as is.
        return v



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from ...configuration.utils import TGConfigError
from ...support.converters import asbool, asint
from ...util.bunch import get_partial_dict
from ..base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigReadyConfigurationAction,
    ConfigurationComponent,
)


class MingConfigurationComponent(ConfigurationComponent):
    """Support MongoDB through the Ming Object Document Mapper.

    Configures the ming connection to MongoDB,
    the automatic session cleanup at the end of each request
    and support for the UnitOfWork flushing/discard based
    on request success or failures.

    The configured Ming session is made available as
    ``tg.config['MingSession']``.

    Options:

        * ``ming.enabled``: Enable Ming in your application
        * ``ming.url``: MongoDB url to connect to.
        * ``ming.db``: Database to use. Can also be specified in ``ming.url``.
        * ``ming.autoflush``: Automatically flush the unit of work as the end
                              of each request, unless there was a failure.
        * ``ming.connection.*``: Options provided to the ming engine to configure
                                 the datastore and the ``MongoClient``. Refer to
                                 :class:`.MongoClient` for available options.

    See :class:`.MingApplicationWrapper` for additional
    configuration options.
    """

    id = "ming"

    def get_defaults(self):
        return {"use_ming": False}

    def get_coercion(self):
        def mongo_read_pref(value):
            from pymongo.read_preferences import ReadPreference

            return getattr(ReadPreference, value)

        return {
            "use_ming": asbool,  # backward compatibility for ming.enabled
            "ming.enabled": asbool,
            "ming.autoflush": asbool,
            "ming.connection.max_pool_size": asint,
            "ming.connection.network_timeout": asint,
            "ming.connection.tz_aware": asbool,
            "ming.connection.safe": asbool,
            "ming.connection.journal": asbool,
            "ming.connection.wtimeout": asint,
            "ming.connection.fsync": asbool,
            "ming.connection.ssl": asbool,
            "ming.connection.read_preference": mongo_read_pref,
            "ming.connection.auto_ensure_indexes": asbool,
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self.configure),
            ConfigReadyConfigurationAction(self.setup),
            AppReadyConfigurationAction(self.add_middleware),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator

        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError("Ming Support only works on an ApplicationConfigurator")

        from ...appwrappers.mingflush import MingApplicationWrapper

        configurator.register_application_wrapper(MingApplicationWrapper, after=True)

    def configure(self, conf, app):
        try:
            autoflush_enabled = conf["ming.autoflush"]
        except KeyError:
            autoflush_enabled = True

        conf.setdefault("ming.enabled", conf.get("use_ming", False))
        conf["ming.autoflush"] = conf["ming.enabled"] and autoflush_enabled

    def setup(self, conf, app):
        """Setup MongoDB database engine using Ming"""
        if not conf["ming.enabled"]:
            return

        datastore = self.create_ming_datastore(conf)
        conf["tg.app_globals"].ming_datastore = datastore

        model = self._get_models(conf)
        ming_session = model.init_model(datastore)
        if ming_session is not None:
            # If init_model returns a specific session, keep it around
            # as the MongoDB Session.
            conf["MingSession"] = ming_session

        if "DBSession" not in conf:
            # If the user hasn't specified a default session, assume
            # he/she uses the default DBSession in model
            conf["DBSession"] = model.DBSession

    def add_middleware(self, conf, app):
        """Set up the ming middleware for the unit of work"""
        if not conf["ming.enabled"]:
            return app

        from ming.odm import ThreadLocalODMSession

        from tg.support.middlewares import MingSessionRemoverMiddleware

        return MingSessionRemoverMiddleware(ThreadLocalODMSession, app)

    def _get_models(self, conf):
        try:
            package_models = conf["package"].model
        except AttributeError:
            package_models = None

        model = conf.get("model", package_models)
        if model is None:
            raise TGConfigError("Ming enabled, but no models provided")

        return model

    @staticmethod
    def create_ming_datastore(conf):
        from ming import create_datastore

        url = conf["ming.url"]
        database = conf.get("ming.db", "")
        try:
            connection_options = get_partial_dict("ming.connection", conf)
        except AttributeError:
            connection_options = {}
        if database and url[-1] != "/":
            url += "/"
        ming_url = url + database
        return create_datastore(ming_url, **connection_options)



# -*- coding: utf-8 -*-
import logging

from tg.configuration.utils import coerce_config
from tg.support.converters import asbool, asint

from ..base import (
    AppReadyConfigurationAction,
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
)

log = logging.getLogger(__name__)


class ErrorReportingConfigurationComponent(ConfigurationComponent):
    """Provides Error reporting through Backlash on TurboGears.

    This is enabled/disabled through the ``debug`` configuration option.
    Currently EMail and Sentry backlash reporters can be enabled.

    All the options available for error reporting are configured
    as ``trace_errors.*`` options in your ``app_cfg`` or ``.ini`` files.

    The available options for **EMail** reporter are:

        - ``trace_errors.enable`` -> Enable or disable error reporting,
          by default is enabled if backlash is available and ``debug=false``
        - ``trace_errors.smtp_server`` -> SMTP Server to connect to for sending emails
        - ``trace_errors.smtp_port`` -> SMTP port to connect to
        - ``trace_errors.from_address`` -> Address sending the error emails
        - ``trace_errors.error_email`` -> Address the error emails should be sent to.
        - ``trace_errors.smtp_username`` -> Username to authenticate on SMTP server.
        - ``trace_errors.smtp_password`` -> Password to authenticate on SMTP server.
        - ``trace_errors.smtp_use_tls`` -> Whenever to enable or not TLS for SMTP.
        - ``trace_errors.error_subject_prefix`` -> Prefix to append to error emails,
          by default ``WebApp Error:`` is prepended.
        - ``trace_errors.dump_request`` -> Whenever to attach a request dump to the email so that
          all request data is provided.
        - ``trace_errors.dump_request_size`` -> Do not dump request if it's bigger than this value,
          useful for uploaded files. By default 50K.
        - ``trace_errors.dump_local_frames`` -> Enable dumping local variables in case of crashes.
        - ``trace_errors.dump_local_frames_count`` -> Dump up to X frames when dumping local variables.
          The default is 2
        - ``trace_errors.reporters`` -> Add custom reporters to error reporting middleware.

    Available options for **Sentry** reporter are:

        - ``trace_errors.sentry_dsn`` -> Sentry instance where to send the errors.

    """

    id = "error_reporting"

    def get_defaults(self):
        return {"debug": False, "trace_errors.enable": True}

    def get_coercion(self):
        return {"debug": asbool}

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_backlash),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def _configure_backlash(self, conf, app):
        trace_errors_config = coerce_config(
            conf,
            "trace_errors.",
            {
                "enable": asbool,
                "smtp_use_tls": asbool,
                "dump_request_size": asint,
                "dump_request": asbool,
                "dump_local_frames": asbool,
                "dump_local_frames_count": asint,
            },
        )

        trace_errors_config.setdefault("debug", conf.get("debug", False))
        trace_errors_config.setdefault("error_subject_prefix", "WebApp Error: ")
        trace_errors_config.setdefault(
            "error_message", "An internal server error occurred"
        )
        conf["tg.errorware"] = trace_errors_config

    def _add_middleware(self, conf, app):
        errorware = conf["tg.errorware"]
        if errorware.get("enable", True) and not asbool(conf.get("debug")):
            reporters = []

            if errorware.get("error_email"):
                from backlash.tracing.reporters.mail import EmailReporter

                reporters.append(EmailReporter(**errorware))

            if errorware.get("sentry_dsn"):
                from backlash.tracing.reporters.sentry import SentryReporter

                reporters.append(SentryReporter(**errorware))

            if errorware.get("reporters", []):
                for reporter in errorware["reporters"]:
                    reporters.append(reporter)

            try:
                import backlash
            except ImportError:  # pragma: no cover
                log.warning(
                    "backlash not installed, email tracebacks won't be available"
                )
            else:
                return backlash.TraceErrorsMiddleware(
                    app, reporters, context_injectors=[_turbogears_backlash_context]
                )
        return app


def _turbogears_backlash_context(environ):
    tgl = environ.get("tg.locals")
    return {"request": getattr(tgl, "request", None)}



# -*- coding: utf-8 -*-
import logging

from tg.support.converters import asbool

from ..base import AppReadyConfigurationAction, ConfigurationComponent

log = logging.getLogger(__name__)


class DebuggerConfigurationComponent(ConfigurationComponent):
    """Enabled Backlash interactive debugger support.

    If debug is enabled, the TurboGears app will be wrapped in
    the BackLash debugger middleware which displays
    interactive debugging sessions when a traceback occurs.

    Supported Options:

        * ``debug``: Whenever to enable or not the interactive debugger.

    Make sure that the interactive debugger is never enable on
    production, or it will be a major security issue as it will
    allow full remote code execution.
    """

    id = "debugger"

    def get_defaults(self):
        return {"debug": False}

    def get_coercion(self):
        return {"debug": asbool}

    def get_actions(self):
        return (AppReadyConfigurationAction(self._add_middleware),)

    def _add_middleware(self, conf, app):
        if asbool(conf.get("debug")):
            try:
                import backlash
            except ImportError:  # pragma: no cover
                log.warning("backlash not installed, debug mode won't be available")
            else:
                return backlash.DebuggedApplication(
                    app, context_injectors=[_turbogears_backlash_context]
                )
        return app


def _turbogears_backlash_context(environ):
    tgl = environ.get("tg.locals")
    return {"request": getattr(tgl, "request", None)}



"""Support modules for TG features.

Support modules implement components that are used by TurboGears
to provide the user level API. The support modules will usually
rely on TurboGears specific objects and modules to provide their
functionality.
"""


class NoDefault(object):
    pass


class EmptyContext(object):
    """
    Noop Python Context, does nothing but can be used
    in a with statement to provide a default context.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



# -*- coding: utf-8 -*-
import logging

from .. import milestones
from ..configuration.utils import DependenciesList
from ..request_local import config as reqlocal_config
from ..support.hooks import hooks
from ..util import Bunch
from ..wsgiapp import TGApp
from .base import (
    AppReadyConfigurationAction,
    Configurator,
    EnvironmentLoadedConfigurationAction,
)

log = logging.getLogger(__name__)


class ApplicationConfigurator(Configurator):
    """A Configurator specialised in creating TurboGears applications.

    The configuration blueprint will be used as configuration foundation
    for every new application built from the Configurator, configuration
    blueprint is merged with deployment configuration loaded from
    config files before creating the application.

    Refer to each registered configuration component for
    configuration options available in the Configurator.

    .. versionadded:: 2.4
    """

    def __init__(self):
        super(ApplicationConfigurator, self).__init__()
        self._application_wrappers = DependenciesList()

    def configure(self, global_conf=None, app_conf=None):
        """Initializes configuration of the application.

        Applies all the blueprints, components defaults, global_conf, app_confg
        and components coercions then returns the resulting configuration.
        Also the configuration is set as the currently active one in the process.

        This is the first step invoked by :meth:`.make_wsgi_app` to create a
        new TurboGears application.
        """
        global_conf = Bunch(global_conf or {})
        app_conf = Bunch(app_conf or {})

        conf = super(ApplicationConfigurator, self).configure(global_conf, app_conf)

        # Application wrapper are made available in the configuration for TGApp use.
        conf["application_wrappers"] = self._application_wrappers

        # Load conf dict into the global config object
        try:
            reqlocal_config.pop_process_config()
        except IndexError:  # pragma: no cover
            log.warning(
                "No global config in place, at least defaults should have been here"
            )
        finally:
            reqlocal_config.push_process_config(conf)

        milestones.config_ready.reach()
        hooks.notify("initialized_config", args=(self, conf))
        return conf

    def setup(self, conf):
        """Setup a TurboGears application environment given a configuration.

        This usually involves any configuration step that requires
        all configuration options to be available before it can be executed.

        It's also the place where ApplicationWrappers and ControllerWrappers
        are resolved.
        """
        super(ApplicationConfigurator, self).setup(conf)
        hooks.notify("config_setup", args=(self, conf))

        # Trigger milestone here so that it gets triggered even when
        # websetup (setup-app command) is performed.
        milestones.environment_loaded.reach()

        for _, step in self._components:
            step._apply(EnvironmentLoadedConfigurationAction, conf)

    @classmethod
    def current(self):
        """Retrieves the current ApplicationConfigurator given that a configuration is in place.

        The current configurator is saved within the configuration itself, so whenever
        a configuration is in place it is possible to get access to the configurator
        and then access any of its componenets::

            ApplicationConfigurator.current().get('componentid')
        """
        configurator = reqlocal_config["tg.configurator"]()
        return configurator

    def _make_app(self, conf, wrap_app=None):
        """Create a tg WSGI application and return it.

        ``conf``
            The application's configuration returned by the
            Configurator after :meth:`.configure` and :meth:`.setup`
            were already applied.

        ``wrap_app``
            a WSGI middleware component which takes the core turbogears
            application and wraps it -- inside all the WSGI-components
            provided by TG. This allows you to work with the
            full environment that your TG application would get before
            anything happens in the application itself.

        """
        app = TGApp(conf)

        hooks.notify("configure_new_app", args=(app,))

        if wrap_app is not None:
            app = wrap_app(app)

        app = hooks.notify_with_value("before_wsgi_middlewares", app)

        for _, step in self._components:
            app = step._apply(AppReadyConfigurationAction, conf, app)

        app = hooks.notify_with_value("after_wsgi_middlewares", app)

        return app

    def make_wsgi_app(self, global_conf=None, app_conf=None, wrap_app=None):
        """Creates a new WSGI TurboGears application with provided configuration."""
        conf = self.configure(global_conf, app_conf)
        self.setup(conf)
        return self._make_app(conf, wrap_app=wrap_app)

    def register_application_wrapper(self, wrapper, after=None):
        """Registers a TurboGears application wrapper.

        Application wrappers are like WSGI middlewares but
        are executed in the context of TurboGears and work
        with abstractions like Request and Respone objects.

        See :class:`tg.appwrappers.base.ApplicationWrapper` for
        complete definition of application wrappers.

        The ``after`` parameter defines their position into the
        wrappers chain. The default value ``None`` means they are
        executed in a middle point, so they run after the TurboGears
        wrappers like :class:`.ErrorPageApplicationWrapper` which
        can intercept their response and return an error page.

        Builtin TurboGears wrappers are usually registered with
        ``after=True`` which means they run furthest away from the
        application itself and can intercept the response of any
        other wrapper.

        Providing ``after=False`` means the wrapper will be registered
        near to the application itself (so wrappers registered at default
        position and with after=True will be able to see its response).

        ``after`` parameter can also accept an *application wrapper class*.
        In such case the registered wrapper will be registered right after
        the specified wrapper and so will be a little further from the
        application then the specified one (can see the response of the
        specified one).

        """
        if milestones.environment_loaded.reached:
            # Wrappers are consumed by TGApp constructor, and all the hooks available
            # after the milestone and that could register new wrappers are actually
            # called after TGApp constructors and so the wrappers wouldn't be applied.
            log.warning(
                "Application Wrapper %s registered after environment loaded "
                "milestone has been reached, the wrapper will be used only "
                "for future TGApp instances.",
                wrapper,
            )

        log.debug("Registering application wrapper: %s", wrapper)
        self._application_wrappers.add(wrapper, after=after)

    def replace_application_wrapper(self, key, wrapper):
        """Replaces a registered application wrapper with another one.

        Note that this only applies to future applications that will be created,
        not to already existing ones.
        """
        self._application_wrappers.replace(key, wrapper)



from tg.configuration.utils import TGConfigError

from .i18n import lazy_ugettext


class _ValidationStatus(object):
    """Current request parameters validation status.

    Keeps track of currently validated values, errors and
    the ValidationIntent that caused the validation process.
    """

    __slots__ = ("errors", "values", "exception", "intent")

    def __init__(self, errors=None, values=None, exception=None, intent=None):
        self.errors = errors or {}
        self.values = values or {}
        self.exception = exception
        self.intent = intent

    @property
    def error_handler(self):
        if self.intent is None:
            return None
        return self.intent.error_handler

    @property
    def chain_validation(self):
        if self.intent is None:
            return False
        return self.intent.chain_validation


class _ValidationIntent(object):
    """Details of validation intention.

    Describes how a validation should happen and how
    errors should be handled. It also performs the
    validation itself on the parameters for a given
    controller method.
    """

    def __init__(self, validators, error_handler, chain_validation):
        self.validators = validators
        self.error_handler = error_handler
        self.chain_validation = chain_validation

    def check(self, config, method, params):
        validators = self.validators
        if not validators:
            return params

        validation_exceptions = tuple(config["validation.exceptions"])
        validation_validators = config["validation.validators"]
        validated_params = {}

        # The validator may be a dictionary, on of the registered validators or
        # any object with a "validate" method.
        if isinstance(validators, dict):
            # TG developers can pass in a dict of param names and validators.
            # They are applied one by one and builds up a new set
            # of validated params.

            errors = {}
            for field, validator in validators.items():
                try:
                    validated_params[field] = validator.to_python(params.get(field))
                except validation_exceptions as inv:
                    # catch individual validation errors into the errors dictionary
                    errors[field] = inv

            # Parameters that don't have validators are returned verbatim
            for param, param_value in params.items():
                if param not in validated_params:
                    validated_params[param] = param_value

            # If there are errors, create a compound validation error based on
            # the errors dictionary, and raise it as an exception
            if errors:
                raise TGValidationError(
                    TGValidationError.make_compound_message(errors),
                    value=params,
                    error_dict=errors,
                )
        elif hasattr(validators, "validate") and getattr(
            self, "needs_controller", False
        ):
            # An object with a "validate" method - call it with the parameters
            validated_params = validators.validate(method, params)
        elif hasattr(validators, "validate"):
            # An object with a "validate" method - call it with the parameters
            validated_params = validators.validate(params)
        else:
            schema_class = validators.__class__
            validation_function = None
            for supported_class in validation_validators:
                if issubclass(schema_class, supported_class):
                    validation_function = validation_validators[supported_class]

            if validation_function is None:
                raise TGConfigError(
                    f"No validation validator function found for: {schema_class}"
                )

            validated_params = validation_function(validators, params)

        return validated_params


class TGValidationError(Exception):
    """Invalid data was encountered during validation.

    The constructor can be passed a short message with
    the reason of the failed validation.
    """

    def __init__(self, msg, value=None, error_dict=None):
        super(TGValidationError, self).__init__(msg)
        self.msg = msg
        self.value = value
        self.error_dict = error_dict

    @classmethod
    def make_compound_message(cls, error_dict):
        return str("\n").join(
            str("%s: %s") % errorinfo for errorinfo in error_dict.items()
        )

    def __str__(self):
        return str(self.msg)


class Convert(object):
    """Applies a conversion function as a validator.

    This is meant to implement simple validation mechanism.

    Any callable can be used for ``func`` as far as it accepts an argument and
    returns the converted object. In case of exceptions the validation
    is considered failed and the ``msg`` parameter is displayed as
    an error.

    A ``default`` value can be provided for values that are missing
    (evaluate to false) which will be used in place of the missing value.

    Example::

        @expose()
        @validate({
            'num': Convert(int, 'Must be a number')
        }, error_handler=insert_number)
        def post_pow2(self, num):
            return str(num*num)
    """

    def __init__(self, func, msg=lazy_ugettext("Invalid"), default=None):
        self._func = func
        self._msg = msg
        self._default = default

    def to_python(self, value, state=None):
        if RequireValue.is_empty(value):
            if self._default is None:
                raise TGValidationError(self._msg, value)
            return self._default

        try:
            return self._func(value)
        except Exception:
            raise TGValidationError(self._msg, value)


class RequireValue(object):
    """Mark a value as required during validation.

    This is meant to be used when a value is required,
    but not conversion needs to happen.
    This is usually common for string values.

    In case conversion has to happen, use :class:`Convert`
    which will already fail if no value is provided.

    Example::

        @expose()
        @validate({
            'num': RequireValue('you must provide a number')
        }, error_handler=insert_number)
        def post_pow2(self, num):
            return str(num*num)
    """

    def __init__(self, msg=lazy_ugettext("Required")):
        self._msg = msg

    @staticmethod
    def is_empty(value):
        return value in (None, "", b"", [], {}, tuple())

    def to_python(self, value, state=None):
        if self.is_empty(value):
            raise TGValidationError(self._msg, value)
        return value



import time

from webtest import TestApp

from tg import FullStackApplicationConfigurator, TGController, expose
from tg.configurator.components.error_reporting import (
    ErrorReportingConfigurationComponent,
)
from tg.configurator.components.slow_requests import SlowRequestsConfigurationComponent


def simple_app(environ, start_response):
    status = '200 OK'
    headers = [('Content-type', 'text/plain')]
    start_response(status, headers)
    return ['HELLO']


class TestErrorReporterConfig(object):

    middleware_name = 'TraceErrorsMiddleware'

    def _make(self, **options):
        step = ErrorReportingConfigurationComponent()
        conf = dict(('trace_errors.%s' % k, v) for k, v in options.items() if k != 'debug')
        if 'debug' in options:
            conf['debug'] = options['debug']
        step._configure_backlash(conf, None)
        return step._add_middleware(conf, simple_app)

    def test_enable_none(self):
        app = self._make()
        assert app.__class__.__name__ == self.middleware_name
        assert not app.reporters

    def test_enable_false(self):
        app = self._make(enable=False)
        assert app.__class__.__name__ != self.middleware_name

    def test_enable_true(self):
        app = self._make(enable=True)
        assert app.__class__.__name__ == self.middleware_name
        assert not app.reporters

    def test_enable_email(self):
        app = self._make(error_email='user@somedomain.com', 
                         smtp_server="fakeserver", 
                         from_address="fakesender")
        assert app.__class__.__name__ == self.middleware_name
        assert any(r.__class__.__name__ == 'EmailReporter'
            for r in app.reporters)

    def test_enable_email_requires_options(self):
        try:
            app = self._make(error_email='user@somedomain.com', 
                             from_address="fakesender")
        except ValueError as e:
            assert "smtp_server" in str(e)
        else:
            assert False, "not raised"

    def test_enable_sentry(self):
        app = self._make(sentry_dsn='http://public:secret@example.com/1')
        assert app.__class__.__name__ == self.middleware_name
        assert any(r.__class__.__name__ == 'SentryReporter'
            for r in app.reporters)

    def test_debug_mode(self):
        app = self._make(debug='on', enable=True,
                         error_email='user@somedomain.com')
        assert app is simple_app

    def test_actually_reports(self):
        class RootController(TGController):
            @expose()
            def index(self):
                return 1/0

        REPORTED_CONTEXT = {}
        class Reporter(object):
            def report(self, traceback):
                REPORTED_CONTEXT.update(traceback.context)

        cfg = FullStackApplicationConfigurator()
        cfg.update_blueprint({
            'debug': False,
            'trace_errors.enable': True,
            'trace_errors.reporters': [
                Reporter()
            ],
            'root_controller': RootController()
        })
        app = TestApp(cfg.make_wsgi_app({}, {}))
        app.get('/', status=404, expect_errors=True)

        assert 'request' in REPORTED_CONTEXT


class TestSlowReqsReporterConfig(object):

    middleware_name = 'TraceSlowRequestsMiddleware'

    def _make(self, **options):
        step = SlowRequestsConfigurationComponent()
        conf = dict(('trace_slowreqs.%s' % k, v) for k, v in options.items() if k != 'debug')
        if 'debug' in options:
            conf['debug'] = options['debug']
        step._configure_backlash(conf, None)
        return step._add_middleware(conf, simple_app)

    def test_disable_all(self):
        app = self._make()
        assert app is simple_app

    def test_enable_without_reporter(self):
        app = self._make(enable=True)
        assert app.__class__.__name__ == self.middleware_name
        assert not app.reporters

    def test_enable_email(self):
        app = self._make(enable=True, error_email='user@somedomain.com',
                         smtp_server="fakesmtp", from_address="fakefrom")
        assert app.__class__.__name__ == self.middleware_name
        assert any(r.__class__.__name__ == 'EmailReporter'
            for r in app.reporters)

    def test_enable_sentry(self):
        app = self._make(enable=True, sentry_dsn='http://public:secret@example.com/1')
        assert app.__class__.__name__ == self.middleware_name
        assert any(r.__class__.__name__ == 'SentryReporter'
            for r in app.reporters)

    def test_debug_mode(self):
        app = self._make(debug='on', enable=True,
                         error_email='user@somedomain.com')
        assert app is simple_app

    def test_backward_compatibility(self):
        conf = {'tg.errorware': {'SOMETHING': True}}

        step = SlowRequestsConfigurationComponent()
        step._configure_backlash(conf, None)

        assert conf['tg.slowreqs']['SOMETHING'] == True

    def test_actually_reports(self):
        class RootController(TGController):
            @expose()
            def index(self):
                # Wait for the slow req reporter to report the request.
                for i in range(100):
                    if 'request' in REPORTED_CONTEXT:
                        break
                    time.sleep(0.01)
                else:
                    assert False, 'Timeout!'
                return 'HI'

        REPORTED_CONTEXT = {}
        class Reporter(object):
            def report(self, traceback):
                REPORTED_CONTEXT.update(traceback.context)

        cfg = FullStackApplicationConfigurator()
        cfg.update_blueprint({
            'debug': False,
            'trace_slowreqs.enable': True,
            'trace_slowreqs.interval': 0,
            'trace_slowreqs.reporters': [
                Reporter()
            ],
            'root_controller': RootController()
        })
        app = TestApp(cfg.make_wsgi_app({}, {}))
        assert app.get('/').text == 'HI'
        assert 'request' in REPORTED_CONTEXT



import datetime
import json

import pytest

from tg import jsonify

try:
    try:
        import sqlite3
    except:
        import pysqlite2
    from sqlalchemy import (
        Column,
        Date,
        DateTime,
        ForeignKey,
        Integer,
        MetaData,
        String,
        Table,
        Time,
        create_engine,
    )
    from sqlalchemy.orm import Session, mapper, registry, relationship

    engine = create_engine("sqlite:///:memory:")
    mapper_registry = registry()
    metadata = mapper_registry.metadata

    test1 = Table('test1', metadata,
        Column('id', Integer, primary_key=True),
        Column('val', String(8)))

    test2 = Table('test2', metadata,
        Column('id', Integer, primary_key=True),
        Column('test1id', Integer, ForeignKey('test1.id')),
        Column('val', String(8)))

    test3 = Table('test3', metadata,
        Column('id', Integer, primary_key=True),
        Column('val', String(8)))

    test4 = Table('test4', metadata,
        Column('id', Integer, primary_key=True),
        Column('val', String(8)))

    test5 = Table('test5', metadata,
        Column('id', Integer, primary_key=True),
        Column('val', String(8)),
        Column('date', DateTime()),
        Column('time', Time()))

    metadata.create_all(engine)

    class MTest2(object):
        pass
    mapper_registry.map_imperatively(MTest2, test2)

    class MTest1(object):
        pass
    mapper_registry.map_imperatively(MTest1, test1, properties={'test2s': relationship(MTest2)})

    class MTest3(object):
        def __json__(self):
            return {'id': self.id, 'val': self.val, 'customized': True}
    mapper_registry.map_imperatively(MTest3, test3)

    class MTest4(object):
        pass
    mapper_registry.map_imperatively(MTest4, test4)

    class MTest5(object):
        pass
    mapper_registry.map_imperatively(MTest5, test5)

    connection = engine.connect()
    connection.execute(test1.insert(), {'id': 1, 'val': 'bob'})
    connection.execute(test2.insert(), {'id': 1, 'test1id': 1, 'val': 'fred'})
    connection.execute(test2.insert(), {'id': 2, 'test1id': 1, 'val': 'alice'})
    connection.execute(test3.insert(), {'id': 1, 'val': 'bob'})
    connection.execute(test4.insert(), {'id': 1, 'val': 'alberto'})
    connection.execute(test5.insert(), {'id': 1, 'val': 'sometime', 'time': datetime.time(21, 20, 19),
                                        'date': datetime.datetime(2016, 12, 11, 10, 9, 8)})
except ImportError:
    from warnings import warn
    warn('SQLAlchemy or PySqlite not installed - cannot run these tests.')
else:
    def teardown_module():
        connection.close()

    def test_saobj():
        s = Session(engine)
        t = s.get(MTest1, 1)
        assert t
        encoded = jsonify.encode(t)
        expected = json.loads('{"id": 1, "val": "bob"}')
        result = json.loads(encoded)
        assert result == expected, encoded

    def test_salist():
        s = Session(engine)
        t = s.get(MTest1, 1)
        encoded = jsonify.encode(dict(results=t.test2s))
        expected = json.loads('''{"results": [{"test1id": 1, "id": 1, "val": "fred"}, {"test1id": 1, "id": 2, "val": "alice"}]}''')
        result = json.loads(encoded)
        assert result == expected, encoded
        
    def test_select_row():
        s = Session(engine)
        t = connection.execute(test1.select())
        encoded = jsonify.encode(dict(results=t))
        expected = json.loads("""{"results": {"count": -1, "rows": [{"count": 1, "rows": {"id": 1, "val": "bob"}}]}}""")
        result = json.loads(encoded)
        assert result == expected, encoded

    def test_select_rows():
        s = Session(engine)
        t = connection.execute(test2.select())
        encoded = jsonify.encode(dict(results=t))
        expected = json.loads("""{"results": {"count": -1, "rows": [{"count": 1, "rows": {"test1id": 1, "id": 1, "val": "fred"}}, {"count": 1, "rows": {"test1id": 1, "id": 2, "val": "alice"}}]}}""")
        result = json.loads(encoded)
        assert result == expected, encoded

    def test_explicit_saobj():
        s = Session(engine)
        t = s.get(MTest3, 1)
        encoded = jsonify.encode(t)
        expected = json.loads('{"id": 1, "val": "bob", "customized": true}')
        result = json.loads(encoded)
        assert result == expected, encoded

    def test_detached_saobj():
        s = Session(engine)
        t = s.get(MTest1, 1)
        # ensure it can be serialized now
        jsonify.encode(t)
        s.expunge(t)

        with pytest.raises(ValueError):
            jsonify.encode(t)

    def test_select_rows_datetime():
        s = Session(engine)
        t = connection.execute(test5.select())
        encoded = jsonify.encode(dict(results=t), encoder=jsonify.JSONEncoder(isodates=True))
        expected = """{"results": {"count": -1, "rows": [{"count": 1, "rows": {"date": "2016-12-11T10:09:08", "id": 1, "val": "sometime", "time": "21:20:19"}}]}}"""
        encoded = json.loads(encoded)
        expected = json.loads(expected)
        assert encoded == expected, encoded



# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
#
# Adapted to TurboGears 2.3
import logging
from string import Template


def asbool(obj):
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in ["true", "yes", "on", "y", "t", "1"]:
            return True
        elif obj in ["false", "no", "off", "n", "f", "0"]:
            return False
        else:
            raise ValueError("String is not true/false: %r" % obj)
    return bool(obj)


def asint(obj):
    try:
        return int(obj)
    except (TypeError, ValueError):
        raise ValueError("Bad integer value: %r" % obj)


def aslist(obj, sep=None, strip=True):
    if isinstance(obj, str):
        lst = obj.split(sep)
        if strip:
            lst = [v.strip() for v in lst]
        return lst
    elif isinstance(obj, (list, tuple)):
        return obj
    elif obj is None:
        return []
    else:
        return [obj]


def astemplate(obj):
    if isinstance(obj, Template):
        return obj

    if not isinstance(obj, str):
        raise ValueError("Templates must be strings")

    return Template(obj)


def aslogger(val):
    if isinstance(val, logging.Logger):
        return val

    if not isinstance(val, str):
        raise ValueError("Logger names must be strings")

    return logging.getLogger(val)



# -*- coding: utf-8 -*-
from logging import getLogger

from ...configuration import milestones
from ..base import (
    BeforeConfigConfigurationAction,
    ConfigurationComponent,
    EnvironmentLoadedConfigurationAction,
)

log = getLogger(__name__)


class DispatchConfigurationComponent(ConfigurationComponent):
    """Provide support for controller wrappers and dispatching configuration.

    This component allows to setup and configure dispatching options
    and controller wrappers.

    The supported options are:

        - ``root_controller``: A root controller instance to use instead of
                               loading it from the application package.
        - ``disable_request_extensions``: Whenever to disable detection of expected
                                          response type based on request path extension
                                          (like ``/index.html`` returns HTML while
                                          ``/index.json`` returns JSON).
        - ``dispatch_path_translator``: A translation function to apply to request paths.
                                        Can be False to disable translation or True to
                                        apply default escaping. Refer to :class:`.DispatchState`
                                        for details.
        - ``ignore_parameters``: Request parameters that should be ignored when dispatching
                                 requests. Refer to :class:`.DispatchState` for details.
        - ``enable_routing_args``: Set routing args in dispatcher state during dispatch and
                                   call ``_setup_wsgiorg_routing_args`` on root controller to
                                   to allow trapping routing arguments.

    Controller wrappers can be registered by using :meth:`.register_controller_wrapper`::

        configurator.get_component('dispatch').register_controller_wrapper(wrapper)

    """

    id = "dispatch"

    def __init__(self):
        super(DispatchConfigurationComponent, self).__init__()
        self._controller_wrappers = []

    def get_defaults(self):
        return {"enable_routing_args": False, "disable_request_extensions": False}

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_explicit_root_controller),
            EnvironmentLoadedConfigurationAction(self._setup_controller_wrappers),
        )

    def _configure_explicit_root_controller(self, conf, app):
        conf["tg.root_controller"] = conf.pop("root_controller", None)

    def _setup_controller_wrappers(self, conf, app):
        # This trashes away the current config['controller_caller']
        # so that the call is idempotent.
        base_controller_caller = _call_controller

        controller_caller = base_controller_caller
        for wrapper in self._controller_wrappers:
            controller_caller = wrapper(controller_caller)

        conf["controller_caller"] = controller_caller

    def register_controller_wrapper(self, wrapper, controller=None):
        """Registers a TurboGears controller wrapper.

        Controller Wrappers are much like a **decorator** applied to
        every controller.
        They receive the next handler in chain and are expected
        to return a new handler that performs whatever it requires
        and then calls the next handler.

        A simple example for a controller wrapper is a simple logging wrapper::

            def controller_wrapper(caller):
                def call(*args, **kw):
                    try:
                        print 'Before handler!'
                        return caller(*args, **kw)
                    finally:
                        print 'After Handler!'
                return call

            dispatch_component.register_controller_wrapper(controller_wrapper)

        It is also possible to register wrappers for a specific controller::

            dispatch_component.register_controller_wrapper(controller_wrapper, controller=RootController.index)
        """
        if milestones.environment_loaded.reached:
            log.warning(
                "Controller Wrapper %s registered after environment loaded "
                "milestone has been reached, the wrapper will be used only "
                "for future TGApp instances.",
                wrapper,
            )

        log.debug(
            "Registering %s controller wrapper for controller: %s",
            wrapper,
            controller or "ALL",
        )

        if controller is None:
            self._controller_wrappers.append(wrapper)
        else:
            from tg.decorators import Decoration

            deco = Decoration.get_decoration(controller)
            deco._register_controller_wrapper(wrapper)


def _call_controller(tg_config, controller, remainder, params):
    return controller(*remainder, **params)



# -*- coding: utf-8 -*-
from tg.appwrappers.errorpage import ErrorPageApplicationWrapper

from ..base import BeforeConfigConfigurationAction, ConfigurationComponent


class ErrorPagesConfigurationComponent(ConfigurationComponent):
    """Provides support for custom error pages.

    This will enable the required parts to show a custom error page
    when a common HTTP error happens in the TurboGears application.

    Refer to :class:`.ErrorPageApplicationWrapper` for supported options.

    When simple authentication backend is set to ``None`` (auth disabled)
    status code ``401`` will always be handled by custom error pages
    even if not specified in ``errorpage.status_codes``.

    """

    id = "error_pages"

    def get_defaults(self):
        return {"errorpage.enabled": True, "errorpage.status_codes": [403, 404]}

    def get_actions(self):
        return (BeforeConfigConfigurationAction(self._configure_error_pages),)

    def on_bind(self, configurator):
        configurator.register_application_wrapper(
            ErrorPageApplicationWrapper, after=True
        )

    def _configure_error_pages(self, conf, app):
        if (
            conf.get("sa_auth.enabled") is False
            and 401 not in conf["errorpage.status_codes"]
        ):
            # If there's no auth backend configured which traps 401
            # responses we redirect those responses to a nicely
            # formatted error page
            conf["errorpage.status_codes"] = list(conf["errorpage.status_codes"]) + [
                401
            ]



"""Defines and initialises what is exposed as tg.config"""

from collections.abc import MutableMapping as DictMixin
from copy import deepcopy

from ..request_local import config as reqlocal_config
from ..util.bunch import Bunch, get_partial_dict


class _DispatchingConfigWrapper(DictMixin):
    """Wrapper for the Dispatching configuration.

    Simple wrapper for the DispatchingConfig object that provides attribute
    style access to the config dictionary.

    This class works by proxying all attribute and dictionary access to
    the underlying DispatchingConfig config object, which is an application local
    proxy that allows for multiple TG2 applications to live
    in the same process simultaneously, but to always get the right
    config data for the application that's requesting them.

    """

    def __init__(self, dict_to_wrap):
        """Initialize the object by passing in config to be wrapped"""
        self.__dict__["config_proxy"] = dict_to_wrap

    def __getitem__(self, key):
        return self.config_proxy.current_conf()[key]

    def __setitem__(self, key, value):
        self.config_proxy.current_conf()[key] = value

    def __getattr__(self, key):
        """Our custom attribute getter.

        Tries to get the attribute off the wrapped object first,
        if that does not work, tries dictionary lookup, and finally
        tries to grab all keys that start with the attribute and
        return sub-dictionaries that can be looked up.

        """
        try:
            return self.config_proxy.__getattribute__(key)
        except AttributeError:
            try:
                return self.config_proxy.current_conf()[key]
            except KeyError:
                return get_partial_dict(key, self.config_proxy.current_conf(), Bunch)

    def __setattr__(self, key, value):
        self.config_proxy.current_conf()[key] = value

    def __delattr__(self, name):
        try:
            del self.config_proxy.current_conf()[name]
        except KeyError:
            raise AttributeError(name)

    def __delitem__(self, key):
        self.__delattr__(key)

    def __len__(self):
        return len(self.config_proxy.current_conf())

    def __iter__(self):
        return iter(self.config_proxy.current_conf())

    def __repr__(self):
        try:
            return repr(self.config_proxy.current_conf())
        except AttributeError:
            return "<TGConfig: missing>"

    def keys(self):
        return self.config_proxy.keys()


def _init_default_global_config():
    defaults = {
        "debug": False,
        "package": None,
        "tg.app_globals": None,
        "tg.strict_tmpl_context": True,
        "i18n.lang": None,
    }

    # Push an empty config so all accesses to config at import time have something
    # to look at and modify. This config will be merged with the app's when it's
    # built in the paste.app_factory entry point.
    reqlocal_config.push_process_config(deepcopy(defaults))


# Create a config object that has attribute style lookup built in.
_init_default_global_config()
config = _DispatchingConfigWrapper(reqlocal_config)
